import re
import flask_babel
import uuid
# from flask_babelex import lazy_gettext as _
from flask import render_template, g, request
from flask_babel import gettext as _
import wtforms as wtf
from flask_wtf import FlaskForm
from flask_bootstrap import SwitchField
from hiddifypanel.panel import hiddify
from flask_classful import FlaskView
from wtforms.validators import ValidationError
# from gettext import gettext as _

from hiddifypanel.models import Domain, DomainType, StrConfig, ConfigEnum, get_hconfigs
from hiddifypanel.database import db
from hiddifypanel.auth import login_required
from hiddifypanel import hutils
from hiddifypanel.models import *


class QuickSetup(FlaskView):
    decorators = [login_required({Role.super_admin})]

    def index(self):
        return render_template(
            'quick_setup.html', lang_form=get_lang_form(),
            form=get_quick_setup_form(),
            ipv4=hutils.network.get_ip_str(4),
            ipv6=hutils.network.get_ip_str(6),
            admin_link=admin_link(),
            show_domain_info=True)

    def post(self):
        if request.args.get('changepw') == "true":
            AdminUser.current_admin_or_owner().uuid = str(uuid.uuid4())
        set_hconfig(ConfigEnum.first_setup, False)
        quick_form = get_quick_setup_form()
        lang_form = get_lang_form()
        if lang_form.lang_submit.data:
            if lang_form.validate_on_submit():
                set_hconfig(ConfigEnum.lang, lang_form.admin_lang.data)
                set_hconfig(ConfigEnum.admin_lang, lang_form.admin_lang.data)
                set_hconfig(ConfigEnum.country, lang_form.country.data)

                flask_babel.refresh()
                hutils.flask.flash((_('quicksetup.setlang.success')), 'success')
            else:
                hutils.flask.flash((_('quicksetup.setlang.error')), 'danger')

            return render_template(
                'quick_setup.html', form=get_quick_setup_form(True),
                lang_form=get_lang_form(),
                admin_link=admin_link(),
                ipv4=hutils.network.get_ip_str(4),
                ipv6=hutils.network.get_ip_str(6),
                show_domain_info=False)

        if quick_form.validate_on_submit():
            Domain.query.filter(Domain.domain == f'{hutils.network.get_ip_str(4)}.sslip.io').delete()
            db.session.add(Domain(domain=quick_form.domain.data.lower(), mode=DomainType.direct))
            set_hconfig(ConfigEnum.block_iran_sites, quick_form.block_iran_sites.data)
            set_hconfig(ConfigEnum.decoy_domain, quick_form.decoy_domain.data)
            # hiddify.bulk_register_configs([
            #     # {"key": ConfigEnum.telegram_enable, "value": quick_form.enable_telegram.data == True},
            #     # {"key": ConfigEnum.vmess_enable, "value": quick_form.enable_vmess.data == True},
            #     # {"key": ConfigEnum.firewall, "value": quick_form.enable_firewall.data == True},
            #     {"key": ConfigEnum.block_iran_sites, "value": quick_form.block_iran_sites.data == True},
            #     # {"key":ConfigEnum.decoy_domain,"value":quick_form.decoy_domain.data}
            # ])

            from .Actions import Actions
            return Actions().reinstall(domain_changed=True)
        else:
            hutils.flask.flash(_('config.validation-error'), 'danger')
        return render_template(
            'quick_setup.html', form=quick_form, lang_form=get_lang_form(True),
            ipv4=hutils.network.get_ip_str(4),
            ipv6=hutils.network.get_ip_str(6),
            admin_link=admin_link(),
            show_domain_info=False)


def get_lang_form(empty=False):
    class LangForm(FlaskForm):
        admin_lang = wtf.SelectField(
            _("config.admin_lang.label"), choices=[("en", _("lang.en")), ("fa", _("lang.fa")), ("pt", _("lang.pt")), ("zh", _("lang.zh")), ("ru", _("lang.ru"))],
            description=_("config.admin_lang.description"),
            default=hconfig(ConfigEnum.admin_lang))
        # lang=wtf.SelectField(_("config.lang.label"),choices=[("en",_("lang.en")),("fa",_("lang.fa"))],description=_("config.lang.description"),default=hconfig(ConfigEnum.lang))
        country = wtf.SelectField(
            _("config.country.label"), choices=[("ir", _("Iran")), ("zh", _("China")), ("other", "Others")],
            description=_("config.country.description"),
            default=hconfig(ConfigEnum.country))
        lang_submit = wtf.SubmitField(_('Submit'))

    return LangForm(None)if empty else LangForm()


def get_quick_setup_form(empty=False):
    def get_used_domains():
        configs = get_hconfigs()
        domains = []
        for c in configs:
            if "domain" in c:
                domains.append(configs[c])
        for d in Domain.query.all():
            domains.append(d.domain)
        return domains

    class QuickSetupForm(FlaskForm):
        domain_regex = "^([A-Za-z0-9\\-\\.]+\\.[a-zA-Z]{2,})$"

        domain_validators = [
            wtf.validators.Regexp(domain_regex, re.IGNORECASE, _("config.Invalid domain")),
            validate_domain,
            wtf.validators.NoneOf([d.domain.lower() for d in Domain.query.all()], _("config.Domain already used")),
            wtf.validators.NoneOf([c.value.lower() for c in StrConfig.query.all() if "fakedomain" in c.key and c.key != ConfigEnum.decoy_domain], _("config.Domain already used"))]
        domain = wtf.StringField(
            _("domain.domain"),
            domain_validators,
            description=_("domain.description"),
            render_kw={
                "class": "ltr",
                "pattern": domain_validators[0].regex.pattern,
                "title": domain_validators[0].message,
                "required": "",
                "placeholder": "sub.domain.com"})
        # enable_telegram = SwitchField(_("config.telegram_enable.label"), description=_("config.telegram_enable.description"), default=hconfig(ConfigEnum.telegram_enable))
        # enable_firewall = SwitchField(_("config.firewall.label"), description=_("config.firewall.description"), default=hconfig(ConfigEnum.firewall))
        block_iran_sites = SwitchField(_("config.block_iran_sites.label"), description=_(
            "config.block_iran_sites.description"), default=hconfig(ConfigEnum.block_iran_sites))
        # enable_vmess = SwitchField(_("config.vmess_enable.label"), description=_("config.vmess_enable.description"), default=hconfig(ConfigEnum.vmess_enable))
        decoy_domain = wtf.StringField(_("config.decoy_domain.label"), description=_("config.decoy_domain.description"), default=hconfig(
            ConfigEnum.decoy_domain), validators=[wtf.validators.Regexp(domain_regex, re.IGNORECASE, _("config.Invalid domain")), hutils.flask.validate_domain_exist])
        submit = wtf.SubmitField(_('Submit'))

    return QuickSetupForm(None) if empty else QuickSetupForm()


def validate_domain(form, field):
    domain = field.data
    dip = hutils.network.get_domain_ip(domain)
    if dip is None:
        raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))

    myip = hutils.network.get_ip(4)
    if dip and myip != dip:
        raise ValidationError(_("Domain (%(domain)s)-> IP=%(domain_ip)s is not matched with your ip=%(server_ip)s which is required in direct mode", server_ip=myip, domain_ip=dip, domain=domain))


def admin_link():
    domains = get_panel_domains()
    return hiddify.get_account_panel_link(g.account, domains[0] if len(domains)else hutils.network.get_ip_str(4))



from flask.views import MethodView
from flask import current_app as app
from loguru import logger

from hiddifypanel.models import Child, Role
from hiddifypanel.auth import login_required

from .schema import ChildStatusInputSchema, ChildStatusOutputSchema


class StatusApi(MethodView):
    decorators = [login_required(node_auth=True)]

    @app.input(ChildStatusInputSchema, arg_name='data')  # type: ignore
    @app.output(ChildStatusOutputSchema)  # type: ignore
    def post(self, data):
        logger.info(f"Checking the existence of child with unique_id: {data['child_unique_id']}")
        res = ChildStatusOutputSchema()
        res.existance = False  # type: ignore

        child = Child.query.filter(Child.unique_id == data['child_unique_id']).first()
        if child:
            logger.info(f"Child with unique_id: {data['child_unique_id']} exists")
            res.existance = True  # type: ignore

        return res



from .ssh_liberty_bridge_api import SSHLibertyBridgeApi
from .xray_api import XrayApi
from .singbox_api import SingboxApi
from .wireguard_api import WireguardApi
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
from collections import defaultdict

drivers = [XrayApi(), SingboxApi(), SSHLibertyBridgeApi(), WireguardApi()]


def enabled_drivers():
    return [d for d in drivers if d.is_enabled()]


def get_users_usage(reset=True):
    res = {}
    users = list(User.query.all())
    res = defaultdict(lambda: {'usage': 0, 'devices': ''})
    for driver in enabled_drivers():
        all_usage = driver.get_all_usage(users)
        for user, usage in all_usage.items():
            if usage:
                res[user]['usage'] += usage
            # res[user]['devices'] +=usage
    return res


def get_enabled_users():
    from collections import defaultdict
    d = defaultdict(int)
    total = 0
    for driver in enabled_drivers():
        try:
            for u, v in driver.get_enabled_users().items():
                if not v:
                    continue
                d[u] += 1
            total += 1
        except Exception as e:
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error in get_enabled users')

    res = defaultdict(bool)
    for u, v in d.items():
        res[u] = v >= total  # ignore singbox
    return res


def add_client(user: User):
    for driver in enabled_drivers():
        try:
            driver.add_client(user)
        except Exception as e:
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error {e} in add client for user={user.uuid}')


def remove_client(user: User):
    for driver in enabled_drivers():
        try:
            driver.remove_client(user)
        except Exception as e:
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error {e} in remove client for user={user.uuid}')



from apiflask import APIBlueprint

bp = APIBlueprint("api_user", __name__, url_prefix="/<proxy_path>/api/v2/user/", enable_openapi=True)

bp_uuid = APIBlueprint("api_user_uuid", __name__, url_prefix="/<proxy_path>/<uuid:secret_uuid>/api/v2/user/", enable_openapi=True)


def init_app(app):
    with app.app_context():
        from .info_api import InfoAPI
        from .apps_api import AppAPI
        from .mtproxies import MTProxiesAPI
        from .configs_api import AllConfigsAPI
        from .short_api import ShortAPI
        from .apps_api import AppAPI

        bp.add_url_rule("/me/", view_func=InfoAPI)
        bp.add_url_rule("/mtproxies/", view_func=MTProxiesAPI)
        bp.add_url_rule("/all-configs/", view_func=AllConfigsAPI)
        bp.add_url_rule("/short/", view_func=ShortAPI)
        bp.add_url_rule('/apps/', view_func=AppAPI)

        bp_uuid.add_url_rule("/me/", view_func=InfoAPI)
        bp_uuid.add_url_rule("/mtproxies/", view_func=MTProxiesAPI)
        bp_uuid.add_url_rule("/all-configs/", view_func=AllConfigsAPI)
        bp_uuid.add_url_rule("/short/", view_func=ShortAPI)
        bp_uuid.add_url_rule('/apps/', view_func=AppAPI)

    app.register_blueprint(bp)
    app.register_blueprint(bp_uuid)



from flask.views import MethodView

from hiddifypanel.auth import login_required
from hiddifypanel.models import Role


class PingPongApi(MethodView):
    decorators = [login_required(roles={Role.super_admin, Role.admin, Role.agent}, node_auth=True)]

    def get(self):
        return {'msg': 'GET: PONG'}

    def post(self):
        return {'msg': 'POST: PONG'}

    def patch(self):
        return {'msg': 'PATCH: PONG'}

    def delete(self):
        return {'msg': 'DELETE: PONG'}

    def put(self):
        return {'msg': 'PUT: PONG'}



import CloudFlare
from hiddifypanel.models import hconfig, ConfigEnum


def add_or_update_domain(domain, ip, dns_type="A", proxied=True):
    res = False
    if hconfig(ConfigEnum.cloudflare):
        cf = CloudFlare.CloudFlare(token=hconfig(ConfigEnum.cloudflare))
        zone = get_zone(cf, domain)
        if zone:
            dns = get_dns(cf, zone, domain)
            data = {'name': domain[:-len(zone['name'])], 'type': dns_type, 'content': ip, 'proxied': proxied}
            if dns:
                cf.zones.dns_records.put(zone['id'], dns['id'], data=data)
            else:
                cf.zones.dns_records.post(zone['id'], data=data)
            res = True
    return res


def get_zone(cf, domain):
    zones = cf.zones.get()
    res = ""
    if zones:
        related_zone = [z for z in zones if z['name'] in domain]
        if len(related_zone):
            res = related_zone[0]
    return res


def get_dns(cf, zone, domain):
    dns_rec = cf.zones.dns_records(zone['id'])
    rel_dns = [d for d in dns_rec if d['name'] == domain]
    res = ""
    if len(rel_dns):
        res = rel_dns[0]
    return res



from __future__ import annotations
import uuid

from sqlalchemy import Column, Integer, String, Enum, text
from enum import auto
from strenum import StrEnum
from flask import g, has_app_context


from hiddifypanel.database import db, db_execute
from sqlalchemy_serializer import SerializerMixin


class ChildMode(StrEnum):
    virtual = auto()
    remote = auto()  # it's child
    parent = auto()

# the child model is node


class Child(db.Model, SerializerMixin):  # type: ignore
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(200), nullable=False, unique=False)
    mode = Column(Enum(ChildMode), nullable=False, default=ChildMode.virtual)
    # ip = db.Column(db.String(200), nullable=False, unique=True)
    unique_id = Column(String(200), nullable=False, default=lambda: str(uuid.uuid4()), unique=True)
    domains = db.relationship('Domain', cascade="all,delete", backref='child')  # type: ignore
    proxies = db.relationship('Proxy', cascade="all,delete", backref='child')  # type: ignore
    boolconfigs = db.relationship('BoolConfig', cascade="all,delete", backref='child')  # type: ignore
    strconfigs = db.relationship('StrConfig', cascade="all,delete", backref='child')  # type: ignore
    dailyusages = db.relationship('DailyUsage', cascade="all,delete", backref='child')  # type: ignore

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "mode": self.mode,
            "unique_id": self.unique_id
        }

    @staticmethod
    def add_or_update(commit=True, **data) -> Child:
        dbchild = Child.query.filter(Child.id == data['id']).first()
        if not dbchild:
            dbchild = Child()
            db.session.add(dbchild)
        dbchild.name = data['name']
        dbchild.mode = data['mode']
        dbchild.unique_id = data['unique_id']
        if commit:
            db.session.commit()
        return dbchild

    @staticmethod
    def bulk_register(childs, commit=True):
        for child in childs:
            Child.add_or_update(commit=False, **child)
        if commit:
            db.session.commit()

    @classmethod
    def by_id(cls, id: int) -> 'Child':
        return Child.query.filter(Child.id == id).first()

    @classmethod
    def by_unique_id(cls, unique_id: str) -> 'Child':
        return Child.query.filter(Child.unique_id == unique_id).first()

    @classmethod
    def current(cls) -> "Child":
        if has_app_context() and hasattr(g, "child"):
            return g.child
        child = Child.by_id(0)
        if child is None:
            tmp_uuid = str(uuid.uuid4())
            db.session.add(Child(id=0, unique_id=tmp_uuid, name="Root"))
            db.session.commit()
            db_execute(f"update child set id=0 where unique_id='{tmp_uuid}'", commit=True)
            child = Child.by_id(0)

        return child

    @classmethod
    @property
    def node(cls) -> "Child | None":
        if has_app_context() and hasattr(g, "node"):
            return g.node



from typing import Optional, Union, Type
from apiflask import Schema, fields
import traceback
import requests
from loguru import logger
from hiddifypanel.models import hconfig, ConfigEnum


class NodeApiErrorSchema(Schema):
    msg = fields.String(required=True)
    stacktrace = fields.String(required=True)
    code = fields.Integer(required=True)
    reason = fields.String(required=True)


class NodeApiClient():
    def __init__(self, base_url: str, apikey: Optional[str] = None, max_retry: int = 3):
        self.base_url = base_url if base_url.endswith('/') else base_url+'/'
        self.max_retry = max_retry
        self.headers = {'Hiddify-API-Key': apikey or hconfig(ConfigEnum.unique_id)}

    def __call(self, method: str, path: str, payload: Optional[Schema], output_schema: Type[Union[Schema, dict]]) -> Union[dict, NodeApiErrorSchema]:  # type: ignore
        retry_count = 1
        full_url = self.base_url + path.removeprefix('/')
        while 1:
            try:
                # TODO: implement it with aiohttp

                logger.trace(f"Attempting {method} request to node at {full_url}")

                # send request
                if payload:
                    response = requests.request(method, full_url, json=payload.dump(payload), headers=self.headers)
                else:
                    response = requests.request(method, full_url, headers=self.headers)

                # parse response
                response.raise_for_status()
                resp = response.json()
                if not resp:
                    err = NodeApiErrorSchema()
                    err.msg = 'Empty response'  # type: ignore
                    err.stacktrace = ''  # type: ignore
                    err.code = response.status_code  # type: ignore
                    err.reason = response.reason  # type: ignore
                    with logger.contextualize(payload=payload):
                        logger.warning(f"Received empty response from {full_url} with method {method}")
                    return err

                logger.trace(f"Successfully received response from {full_url}")
                return resp if isinstance(output_schema, type(dict)) else output_schema().load(resp)  # type: ignore

            except requests.HTTPError as e:
                if retry_count >= self.max_retry:
                    stack_trace = traceback.format_exc()
                    err = NodeApiErrorSchema()
                    err.msg = str(e)  # type: ignore
                    err.stacktrace = stack_trace  # type: ignore
                    err.code = response.status_code  # type: ignore
                    err.reason = response.reason  # type: ignore
                    with logger.contextualize(status_code=err.code, reason=err.reason, stack_trace=stack_trace, payload=payload):
                        logger.error(f"HTTP error after {self.max_retry} retries")
                        logger.exception(e)
                    return err

                logger.warning(f"Error occurred: {e} from {full_url} with method {method}, retrying... ({retry_count}/{self.max_retry})")
                retry_count += 1

    def get(self, path: str, output: Type[Union[Schema, dict]]) -> Union[dict, NodeApiErrorSchema]:
        return self.__call("GET", path, None, output)

    def post(self, path: str, payload: Optional[Schema], output: Type[Union[Schema, dict]]) -> Union[dict, NodeApiErrorSchema]:
        return self.__call("POST", path, payload, output)

    def put(self, path: str, payload: Optional[Schema], output: Type[Union[Schema, dict]]) -> Union[dict, NodeApiErrorSchema]:
        return self.__call("PUT", path, payload, output)



import json
import os

from .abstract_driver import DriverABS
from hiddifypanel.models import User, hconfig, ConfigEnum
from hiddifypanel.panel.run_commander import Command, commander


class WireguardApi(DriverABS):
    def is_enabled(self) -> bool:
        return hconfig(ConfigEnum.wireguard_enable)
    WG_LOCAL_USAGE_FILE_PATH = './hiddify_usages.json'

    def __init__(self) -> None:
        super().__init__()
        # create empty local usage file
        if not os.path.isfile(WireguardApi.WG_LOCAL_USAGE_FILE_PATH):
            with open(WireguardApi.WG_LOCAL_USAGE_FILE_PATH, 'w+') as f:
                json.dump({}, f)

    def __get_wg_usages(self) -> dict:
        raw_output = commander(Command.update_wg_usage, run_in_background=False)
        data = {}
        for line in raw_output.split('\n'):
            if not line:
                continue
            sections = line.split()
            if len(sections) < 3:
                continue
            data[sections[0]] = {
                'down': int(sections[1]),
                'up': int(sections[2]),
            }
        return data

    def __get_local_usage(self) -> dict:

        with open(WireguardApi.WG_LOCAL_USAGE_FILE_PATH, 'r') as f:
            data = json.load(f)
            return data

    def __sync_local_usages(self) -> dict:
        local_usage = self.__get_local_usage()
        wg_usage = self.__get_wg_usages()
        res = {}
        # remove local usage that is removed from wg usage
        for local_wg_pub in local_usage.copy().keys():
            if local_wg_pub not in wg_usage:
                del local_usage[local_wg_pub]

        for wg_pub, wg_usage in wg_usage.items():
            if not local_usage.get(wg_pub):
                local_usage[wg_pub] = wg_usage
                continue
            res[wg_pub] = self.calculate_reset(local_usage[wg_pub], wg_usage)
            local_usage[wg_pub] = wg_usage

        with open(WireguardApi.WG_LOCAL_USAGE_FILE_PATH, 'w') as f:
            json.dump(local_usage, f)
        return res

    def calculate_reset(self, last_usage: dict, current_usage: dict) -> dict:
        res = {
            'up': current_usage['up'] - last_usage['up'],
            'down': current_usage['down'] - last_usage['down'],
        }

        if res['up'] < 0:
            res['up'] = 0
        if res['down'] < 0:
            res['down'] = 0
        return res

    def get_enabled_users(self):
        if not hconfig(ConfigEnum.wireguard_enable):
            return {}
        usages = self.__get_wg_usages()
        wg_pubs = set(usages.keys())

        users = User.query.all()
        enabled = {}
        for u in users:
            if u.wg_pub in wg_pubs:
                enabled[u.uuid] = 1
            else:
                enabled[u.uuid] = 0
        return enabled

    def add_client(self, user):
        pass

    def remove_client(self, user):
        pass

    def get_all_usage(self, users, reset=True):
        if not hconfig(ConfigEnum.wireguard_enable):
            return {}
        all_usages = self.__sync_local_usages()
        res = {}
        for u in users:
            if use := all_usages.get(u.wg_pub):
                res[u] = use['up'] + use['down']
            else:
                res[u] = 0
        return res



import datetime
import uuid
import json
import os
import click
from dateutil import relativedelta


from hiddifypanel import hutils

from hiddifypanel.models import *
from hiddifypanel.panel import hiddify, usage
from hiddifypanel.database import db
from hiddifypanel.panel.init_db import init_db


def drop_db():
    """Cleans database"""
    db.drop_all()


def downgrade():
    if (hconfig(ConfigEnum.db_version) >= "49"):
        set_hconfig(ConfigEnum.db_version, '42', commit=False)
        StrConfig.query.filter(StrConfig.key.in_([ConfigEnum.tuic_enable, ConfigEnum.tuic_port, ConfigEnum.hysteria_enable,
                               ConfigEnum.hysteria_port, ConfigEnum.ssh_server_enable, ConfigEnum.ssh_server_port, ConfigEnum.ssh_server_redis_url])).delete()
        Proxy.query.filter(Proxy.l3.in_([ProxyL3.ssh, ProxyL3.h3_quic, ProxyL3.custom])).delete()
        db.session.commit()
        os.rename("/opt/hiddify-manager/hiddify-panel/hiddifypanel.db.old", "/opt/hiddify-manager/hiddify-panel/hiddifypanel.db")


def backup():
    dbdict = hiddify.dump_db_to_dict()
    os.makedirs('backup', exist_ok=True)
    dst = f'backup/{datetime.datetime.now().strftime("%Y_%m_%d__%H_%M_%S")}.json'
    with open(dst, 'w') as fp:
        json.dump(dbdict, fp, indent=4, sort_keys=True, default=str)

    if hconfig(ConfigEnum.telegram_bot_token):
        from hiddifypanel.panel.commercial.telegrambot import bot, register_bot
        if not bot.token:
            register_bot(True)

        with open(dst, 'rb') as document:
            for admin in AdminUser.query.filter(AdminUser.mode == AdminMode.super_admin, AdminUser.telegram_id is not None).all():
                caption = ("Backup \n" + admin_links())
                bot.send_document(admin.telegram_id, document, visible_file_name=dst.replace("backup/", ""), caption=caption[:min(len(caption), 1000)])


def all_configs():
    valid_users = [u.to_dict(dump_id=True) for u in User.query.filter((User.usage_limit > User.current_usage)).all() if u.is_active]
    host_child_ids = [c.id for c in Child.query.filter(Child.mode == ChildMode.virtual).all()]
    configs = {
        "users": valid_users,
        "domains": [u.to_dict(dump_ports=True, dump_child_id=True) for u in Domain.query.filter(Domain.child_id.in_(host_child_ids)).all() if "*" not in u.domain],
        # "parent_domains": [hiddify.parent_domain_dict(u) for u in ParentDomain.query.all()],
        # "hconfigs": get_hconfigs(json=True),
        "chconfigs": get_hconfigs_childs(host_child_ids, json=True)
    }

    def_user = None if len(User.query.all()) > 1 else User.query.filter(User.name == 'default').first()
    domains = Domain.query.all()
    sslip_domains = [d.domain for d in domains if "sslip.io" in d.domain]

    configs['chconfigs'][0]['first_setup'] = def_user is not None and len(sslip_domains) > 0
    server_ip = hutils.network.get_ip_str(4)
    owner = AdminUser.get_super_admin()

    configs['admin_path'] = hiddify.get_account_panel_link(owner, server_ip, is_https=False, prefere_path_only=True)
    configs['panel_links'] = []
    configs['panel_links'].append(hiddify.get_account_panel_link(owner, server_ip, is_https=False))
    configs['panel_links'].append(hiddify.get_account_panel_link(owner, server_ip))
    domains = get_panel_domains()

    for d in domains:
        configs['panel_links'].append(hiddify.get_account_panel_link(owner, d.domain))

    print(json.dumps(configs, indent=4))


def update_usage():
    print(usage.update_local_usage())


def test():
    print(ConfigEnum("auto_update1"))


def admin_links():

    server_ip = hutils.network.get_ip_str(4)
    owner = AdminUser.get_super_admin()

    admin_links = f"Not Secure (do not use it - only if others not work):\n   {hiddify.get_account_panel_link(owner, server_ip,is_https=True)}\n"

    domains = get_panel_domains()
    admin_links += f"Secure:\n"
    if not any([d for d in domains if 'sslip.io' not in d.domain]):
        admin_links += f"   (not signed) {hiddify.get_account_panel_link(owner, server_ip)}\n"

    for d in domains:
        admin_links += f"   {hiddify.get_account_panel_link(owner, d.domain)}\n"

    print(admin_links)
    return admin_links


def admin_path():
    admin = AdminUser.get_super_admin()
    # WTF is the owner and server_id?
    domain = get_panel_domains()[0]
    print(hiddify.get_account_panel_link(admin, domain, prefere_path_only=True))


# def get_this_host_domains():
#     current_child_ids


def hysteria_domain_port():
    if not hconfig(ConfigEnum.hysteria_enable):
        return
    out = []
    for i, domain in enumerate(Domain.query.filter(Domain.mode.in_([DomainType.direct, DomainType.relay, DomainType.fake])).all()):
        out.append(f"{domain.domain}:{int(hconfig(ConfigEnum.hysteria_port))+domain.id}")
    print(";".join(out))


def tuic_domain_port():
    if not hconfig(ConfigEnum.tuic_enable):
        return
    out = []
    for i, domain in enumerate(Domain.query.filter(Domain.mode.in_([DomainType.direct, DomainType.relay, DomainType.fake])).all()):
        out.append(f"{domain}:{int(hconfig(ConfigEnum.tuic_port))+domain.id}")
    print(";".join(out))


def init_app(app):
    # add multiple commands in a bulk
    # print(app.config['SQLALCHEMY_DATABASE_URI'] )
    for command in [hysteria_domain_port, tuic_domain_port, init_db, drop_db, all_configs, update_usage, test, admin_links, admin_path, backup, downgrade]:
        app.cli.add_command(app.cli.command()(command))

    @ app.cli.command()
    @ click.option("--domain", "-d")
    def add_domain(domain):
        # TODO: Fix this
        if Domain.query.filter(Domain.domain == domain).first():
            return "Domain already exist."
        d = Domain(domain=domain)
        if not hutils.node.is_parent():
            d.mode = DomainType.direct
        db.session.add(d)
        db.session.commit()
        return "success"

    @ app.cli.command()
    @ click.option("--admin_secret", "-a")
    def set_admin_secret(admin_secret):
        StrConfig.query.filter(StrConfig.key == ConfigEnum.admin_secret).update({'value': admin_secret})
        db.session.commit()
        return "success"

    @ app.cli.command()
    @ click.option("--key", "-k")
    @ click.option("--val", "-v")
    def set_setting(key, val):
        old_hconfigs = get_hconfigs()
        hiddify.add_or_update_config(key=key, value=val)

        return "success"

    @ app.cli.command()
    @ click.option("--config", "-c")
    def import_config(config):
        next10year = datetime.date.today() + relativedelta.relativedelta(years=10)
        data = []
        if "USER_SECRET" in config:
            secrets = config["USER_SECRET"].split(";")
            for i, s in enumerate(secrets):
                data.append(User(name=f"default {i}", uuid=uuid.UUID(s), usage_limit_GB=9000, package_days=3650))

        if "MAIN_DOMAIN" in config:
            domains = config["MAIN_DOMAIN"].split(";")
            for i, d in enumerate(domains):
                if not Domain.query.filter(Domain.domain == d).first():
                    data.append(Domain(domain=d, mode=DomainType.direct),)

        strmap = {
            "TELEGRAM_FAKE_TLS_DOMAIN": ConfigEnum.telegram_fakedomain,
            "TELEGRAM_SECRET": ConfigEnum.shared_secret,
            "SS_FAKE_TLS_DOMAIN": ConfigEnum.ssfaketls_fakedomain,
            "FAKE_CDN_DOMAIN": ConfigEnum.domain_fronting_domain,
            "BASE_PROXY_PATH": ConfigEnum.proxy_path,
            "ADMIN_SECRET": ConfigEnum.admin_secret,
            "TELEGRAM_AD_TAG": ConfigEnum.telegram_adtag
        }
        boolmap = {
            "ENABLE_SS": ConfigEnum.ssfaketls_enable,
            "ENABLE_TELEGRAM": ConfigEnum.telegram_enable,
            "ENABLE_VMESS": ConfigEnum.vmess_enable,
            # "ENABLE_MONITORING":ConfigEnum.ssfaketls_enable,
            "ENABLE_FIREWALL": ConfigEnum.firewall,
            "ENABLE_NETDATA": ConfigEnum.netdata,
            "ENABLE_HTTP_PROXY": ConfigEnum.http_proxy_enable,
            "ALLOW_ALL_SNI_TO_USE_PROXY": ConfigEnum.allow_invalid_sni,
            "ENABLE_AUTO_UPDATE": ConfigEnum.auto_update,
            "ENABLE_SPEED_TEST": ConfigEnum.speed_test,
            "BLOCK_IR_SITES": ConfigEnum.block_iran_sites,
            "ONLY_IPV4": ConfigEnum.only_ipv4
        }

        for k in config:
            if k in strmap:
                if hconfig(strmap[k]) is None:
                    data.append(StrConfig(key=strmap[k], value=config[k]))
                else:
                    StrConfig.query.filter(StrConfig.key == strmap[k]).update({
                        'value': config[k]
                    })
            if k in boolmap:
                if hconfig(boolmap[k]) is None:
                    data.append(BoolConfig(key=boolmap[k], value=config[k]))
                else:
                    BoolConfig.query.filter(BoolConfig.key == strmap[k]).update({
                        'value': config[k]
                    })
        if len(data):
            db.session.bulk_save_objects(data)
        db.session.commit()

    @ app.cli.command()
    @ click.option("--xui_db_path", "-x")
    def xui_importer(xui_db_path):
        try:
            hutils.importer.xui.import_data(xui_db_path)
            print('success')
        except Exception as e:
            print(f'failed to import xui data: Error: {e}')



from typing import Set
from apiflask import HTTPBasicAuth, HTTPTokenAuth
from flask_httpauth import MultiAuth

from hiddifypanel.models.user import User
from hiddifypanel.models.admin import AdminUser
from flask import session
from strenum import StrEnum

basic_auth = HTTPBasicAuth()
api_auth = HTTPTokenAuth("ApiKey")
multi_auth = MultiAuth(basic_auth, api_auth)


class AccountRole(StrEnum):
    user = 'user'
    admin = 'admin'
    super_admin = 'super_admin'
    agent = 'agent'


@basic_auth.verify_password
def verify_basic_auth_password(username, password) -> AdminUser | User | None:
    # username = username.strip()
    # password = password.strip()
    if username and password:
        return User.query.filter(
            User.username == username, User.password == password).first() or AdminUser.query.filter(
            AdminUser.username == username, AdminUser.password == password).first()


@api_auth.verify_token
def verify_api_auth_token(token) -> User | AdminUser | None:
    # for now, token is the same as uuid
    # token = token.strip()
    if token:
        return User.by_uuid(token) or AdminUser.by_uuid(token)

    # we dont' set session for api auth
    # if check_session:Admin
    #     account = verify_from_session(roles)
    #     if account:
    #         return account


def set_admin_authentication_in_session(admin: AdminUser) -> None:
    if not session.get('account'):
        session['account'] = {
            'uuid': admin.uuid,
            'role': get_account_role(admin),
            # 'username': res.username,
        }


def set_user_authentication_in_session(user: User) -> None:
    session['user_sign'] = {'uuid': user.uuid}


def verify_admin_authentication_from_session() -> User | AdminUser | None:
    if session.get('account'):
        return User.by_uuid(session['account']['uuid']) or AdminUser.by_uuid(session['account']['uuid'])


def verify_user_authentication_from_session() -> User | AdminUser | None:
    if session.get('user_sign'):
        return User.by_uuid(session['user_sign']['uuid'])


# actually this is not used, we authenticate the client and setup it in
# the flask.g object, then in views we re-authenticate with their roles to
# see if they have access to the view or not
@api_auth.get_user_roles
@basic_auth.get_user_roles
def get_account_role(account) -> AccountRole | None:
    '''Returns user/admin role
     Allowed roles are:
     - for user:
        - user
     - for admin:
        - super_admin
        - admin
        - agent
    '''
    if isinstance(account, User):
        return AccountRole.user
    if isinstance(account, AdminUser):
        match account.mode:
            case 'super_admin':
                return AccountRole.super_admin
            case 'admin':
                return AccountRole.admin
            case 'agent':
                return AccountRole.agent


def standalone_admin_basic_auth_verification() -> AdminUser | None:
    auth = basic_auth.get_auth()

    if auth:
        account = verify_basic_auth_password(auth.username, auth.password)
        if account and isinstance(account, AdminUser):
            set_admin_authentication_in_session(account)
            return account

    return verify_admin_authentication_from_session()


def standalone_user_basic_auth_verification() -> User | None:
    auth = basic_auth.get_auth()

    if auth and auth.username and auth.password:
        user = verify_basic_auth_password(auth.username, auth.password)
        if user and isinstance(user, User):
            set_user_authentication_in_session(user)
            return user
    return verify_user_authentication_from_session()


def standalone_api_auth_verify():
    auth = api_auth.get_auth()
    try:
        if hasattr(auth, 'token'):
            account = verify_api_auth_token(auth.token)
            if account:
                return account
    except AttributeError:
        return None


# ADD ERROR HANDLING TO AUTHENTICATIONS



from .abstract_driver import DriverABS
from hiddifypanel.models import *
import redis

USERS_SET = "ssh-server:users"
USERS_USAGE = "ssh-server:users-usage"


class SSHLibertyBridgeApi(DriverABS):
    def is_enabled(self) -> bool:
        return hconfig(ConfigEnum.ssh_server_enable)

    def get_ssh_redis_client(self):
        if not hasattr(self, 'redis_client'):
            self.redis_client = redis.from_url('unix:///opt/hiddify-manager/other/redis/run.sock?db=1', decode_responses=True)

        return self.redis_client

    def get_enabled_users(self):
        redis_client = self.get_ssh_redis_client()
        members = redis_client.smembers(USERS_SET)
        return {m.split("::")[0]: 1 for m in members}

    def add_client(self, user):
        print(f'Adding SSH {user}')
        redis_client = self.get_ssh_redis_client()
        redis_client.sadd(USERS_SET, f'{user.uuid}::{user.ed25519_public_key}')
        redis_client.save()

    def remove_client(self, user):
        redis_client = self.get_ssh_redis_client()
        redis_client.srem(USERS_SET, f'{user.uuid}::{user.ed25519_public_key}')
        redis_client.hdel(USERS_USAGE, f'{user.uuid}')
        redis_client.save()

    def get_all_usage(self, users):
        return {u: self.get_usage_imp(u.uuid) for u in users}

    def get_usage_imp(self, client_uuid: str, reset: bool = True) -> int:
        redis_client = self.get_ssh_redis_client()
        value = redis_client.hget(USERS_USAGE, client_uuid)

        if value is None:
            return 0

        value = int(value)

        if reset:
            redis_client.hincrby(USERS_USAGE, client_uuid, -value)
            redis_client.save()
        if value:
            print(f'ssh usage {client_uuid} {value}')
        return value



from apiflask import fields, Schema
from marshmallow import ValidationError

from hiddifypanel.models import DomainType, ProxyProto, ProxyL3, ProxyTransport, ProxyCDN, ConfigEnum, ChildMode
from hiddifypanel.panel.commercial.restapi.v2.admin.schema import UserSchema, AdminSchema


def hconfig_key_validator(value):
    if value not in [c.name for c in ConfigEnum]:
        raise ValidationError(f"{value} is not a valid hconfig key.")
    return value


class DomainSchema(Schema):
    child_unique_id = fields.String(description="The child's unique id")
    domain = fields.String(required=True, description="The domain name")
    alias = fields.String(description="The domain alias", allow_none=True)
    sub_link_only = fields.Boolean(required=True, description="Is the domain sub link only")
    mode = fields.Enum(DomainType, required=True, description="The domain type")
    cdn_ip = fields.String(description="The cdn ip", allow_none=True)
    grpc = fields.Boolean(required=True, description="Is the domain grpc")
    servernames = fields.String(description="The servernames", allow_none=True)
    show_domains = fields.List(fields.String(), desciption="The list of domains to show")


class ProxySchema(Schema):
    child_unique_id = fields.String(description="The child's unique id")
    name = fields.String(required=True, description="The proxy name")
    enable = fields.Boolean(required=True, description="Is the proxy enabled")
    proto = fields.Enum(ProxyProto, required=True, description="The proxy protocol")
    l3 = fields.Enum(ProxyL3, required=True, description="The proxy l3")
    transport = fields.Enum(ProxyTransport, required=True, description="The proxy transport")
    cdn = fields.Enum(ProxyCDN, required=True, description="The proxy cdn")


class StringOrBooleanField(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, (str, bool)):
            return str(value)
        else:
            raise ValidationError("Value must be a string or a boolean.")

    def _serialize(self, value, attr, obj, **kwargs):
        return value


class HConfigSchema(Schema):
    child_unique_id = fields.String(description="The child's unique id")
    key = fields.String(required=True, description="The config key", validate=hconfig_key_validator)  # type: ignore
    value = StringOrBooleanField(required=True, description="The config value")


# region usage
class UsageData(Schema):
    uuid = fields.UUID(required=True, desciption="The user uuid")
    usage = fields.Integer(required=True, description="The user usage in bytes")
    devices = fields.List(fields.String(required=True, description="The user connected devices"))


class UsageInputOutputSchema(Schema):
    usages = fields.List(fields.Nested(UsageData), required=True, description="The list of usages")
# endregion


# region sync
class SyncInputSchema(Schema):
    # users = fields.List(fields.Nested(UserSchema),required=True,description="The list of users")
    domains = fields.List(fields.Nested(DomainSchema), required=True, description="The list of domains")
    proxies = fields.List(fields.Nested(ProxySchema), required=True, description="The list of proxies")
    # parent_domains = fields.List(fields.Nested(ParentDomainSchema),required=True,description="The list of parent domains")
    # admin_users = fields.List(fields.Nested(AdminSchema),required=True,description="The list of admin users")
    hconfigs = fields.List(fields.Nested(HConfigSchema), required=True, description="The list of configs")


class SyncOutputSchema(Schema):
    users = fields.List(fields.Nested(UserSchema), required=True, description="The list of users")
    admin_users = fields.List(fields.Nested(AdminSchema), required=True, description="The list of admin users")

# endregion


# region child status
class ChildStatusInputSchema(Schema):
    child_unique_id = fields.String(required=True, description="The child's unique id")


class ChildStatusOutputSchema(Schema):
    existance = fields.Boolean(required=True, description="Whether child exists")

# end region


# region register

class RegisterDataSchema(Schema):
    users = fields.List(fields.Nested(UserSchema), required=True, description="The list of users")
    domains = fields.List(fields.Nested(DomainSchema), required=True, description="The list of domains")
    proxies = fields.List(fields.Nested(ProxySchema), required=True, description="The list of proxies")
    admin_users = fields.List(fields.Nested(AdminSchema), required=True, description="The list of admin users")
    hconfigs = fields.List(fields.Nested(HConfigSchema), required=True, description="The list of configs")


class RegisterInputSchema(Schema):
    panel_data = fields.Nested(RegisterDataSchema, required=True, description="The child's data")
    unique_id = fields.String(required=True, description="The child's unique id")
    name = fields.String(required=True, description="The child's name")
    mode = fields.Enum(ChildMode, required=True, description="The child's mode")


class RegisterOutputSchema(Schema):
    parent_unique_id = fields.String(description="The parent's unique id")
    users = fields.List(fields.Nested(UserSchema), required=True, description="The list of users")
    admin_users = fields.List(fields.Nested(AdminSchema), required=True, description="The list of admin users")

# endregion



from loguru import logger

from hiddifypanel.models import hconfig, ConfigEnum, PanelMode, User
from hiddifypanel.cache import cache
from hiddifypanel.panel.commercial.restapi.v2.parent.schema import UsageInputOutputSchema, UsageData
from hiddifypanel.panel.commercial.restapi.v2.panel.schema import PanelInfoOutputSchema
from .api_client import NodeApiClient, NodeApiErrorSchema


def is_child() -> bool:
    return hconfig(ConfigEnum.panel_mode) == PanelMode.child


def is_parent() -> bool:
    return hconfig(ConfigEnum.panel_mode) == PanelMode.parent

# region usage


def get_users_usage_data_for_api() -> UsageInputOutputSchema:
    res = UsageInputOutputSchema()
    res.usages = []  # type: ignore
    for u in User.query.all():
        usage_data = UsageData()
        usage_data.uuid = u.uuid
        usage_data.usage = u.current_usage
        usage_data.devices = u.devices
        res.usages.append(usage_data)  # type: ignore
    return res


def convert_usage_api_response_to_dict(data: dict) -> dict:
    converted = {}
    for i in data['usages']:  # type: ignore
        converted[str(i['uuid'])] = {
            'usage': i['usage'],
            'devices': ','.join(i['devices'])  # type: ignore
        }
    return converted

# endregion


#@cache.cache(ttl=150)
def is_panel_active(domain: str, proxy_path: str,apikey:str|None = None) -> bool:
    base_url = f'https://{domain}/{proxy_path}'
    res = NodeApiClient(base_url,apikey).get('/api/v2/panel/ping/', dict)
    if isinstance(res, NodeApiErrorSchema):
        logger.error(f"Error while checking if panel is active: {res.msg}")
        return False
    if 'PONG' in res['msg']:
        logger.debug(f"Panel is active: {res['msg']}")
        return True
    logger.debug("Panel is not active")
    return False


#@cache.cache(300)
def get_panel_info(domain: str, proxy_path: str,apikey:str|None = None) -> dict | None:
    base_url = f'https://{domain}/{proxy_path}'
    res = NodeApiClient(base_url,apikey).get('/api/v2/panel/info/', PanelInfoOutputSchema)
    if isinstance(res, NodeApiErrorSchema):
        logger.error(f"Error while getting panel info from {domain}: {res.msg}")
        return None
    return res



import xtlsapi
from hiddifypanel.models import *
from .abstract_driver import DriverABS
from collections import defaultdict


class XrayApi(DriverABS):
    def is_enabled(self) -> bool:
        return hconfig(ConfigEnum.core_type) == "xray"

    def get_xray_client(self):
        return xtlsapi.XrayClient('127.0.0.1', 10085)

    def get_enabled_users(self):
        xray_client = self.get_xray_client()
        usages = xray_client.stats_query('user', reset=True)
        res = defaultdict(int)
        for use in usages:
            if "user>>>" not in use.name:
                continue
            uuid = use.name.split(">>>")[1].split("@")[0]
            res[uuid] = 1
        return res

        # xray_client = self.get_xray_client()
        # users = User.query.all()
        # t = "xtls"
        # protocol = "vless"
        # enabled = {}
        # for u in users:
        #     uuid = u.uuid
        #     try:
        #         xray_client.add_client(t, f'{uuid}', f'{uuid}@hiddify.com', protocol=protocol, flow='xtls-rprx-vision', alter_id=0, cipher='chacha20_poly1305')
        #         xray_client.remove_client(t, f'{uuid}@hiddify.com')
        #         enabled[uuid] = 0
        #     except xtlsapi.xtlsapi.exceptions.EmailAlreadyExists as e:
        #         enabled[uuid] = 1
        #     except Exception as e:
        #         print(f"error {e}")
        #         enabled[uuid] = e
        # return enabled

    def get_inbound_tags(self):
        try:
            xray_client = self.get_xray_client()
            inbounds = [inb.name.split(">>>")[1] for inb in xray_client.stats_query('inbound')]
            print(f"Success in get inbound tags {inbounds}")
        except Exception as e:
            print(f"error in get inbound tags {e}")
            inbounds = []
        return list(set(inbounds))

    def add_client(self, user):
        uuid = user.uuid
        xray_client = self.get_xray_client()
        tags = self.get_inbound_tags()
        proto_map = {
            'vless': 'vless',
            'realityin': 'vless',
            'xtls': 'vless',
            'quic': 'vless',
            'trojan': 'trojan',
            'vmess': 'vmess',
            'ss': 'shadowsocks',
            'v2ray': 'shadowsocks',
            'kcp': 'vless',
            'dispatcher': 'trojan',
        }

        def proto(t):
            res = '', ''
            for p, protocol in proto_map.items():
                if p in t:
                    res = p, protocol
                    break
            return res
        for t in tags:
            try:
                p, protocol = proto(t)
                if not p:
                    continue
                if (protocol == "vless" and p != "xtls" and p != "realityin") or "realityingrpc" in t:
                    xray_client.add_client(t, f'{uuid}', f'{uuid}@hiddify.com', protocol=protocol, flow='\0',)
                else:
                    xray_client.add_client(t, f'{uuid}', f'{uuid}@hiddify.com', protocol=protocol,
                                           flow='xtls-rprx-vision', alter_id=0, cipher='chacha20_poly1305')
                # print(f"Success add  {uuid} {t}")
            except Exception as e:
                # print(f"error in add  {uuid} {t} {e}")
                pass

    def remove_client(self, user):
        return self._remove_client(user.uuid)

    def _remove_client(self, uuid):
        xray_client = self.get_xray_client()
        tags = self.get_inbound_tags()

        for t in tags:
            try:
                xray_client.remove_client(t, f'{uuid}@hiddify.com')
                print(f"Success remove  {uuid} {t}")
            except Exception as e:
                print(f"error in remove  {uuid} {t} {e}")
                pass

    def get_all_usage(self, users):
        xray_client = self.get_xray_client()
        usages = xray_client.stats_query('user', reset=True)
        uuid_user_map = {u.uuid: u for u in users}
        res = defaultdict(int)
        for use in usages:
            if "user>>>" not in use.name:
                continue
            uuid = use.name.split(">>>")[1].split("@")[0]
            if u := uuid_user_map.get(uuid):
                res[u] += use.value
            else:
                self._remove_client(uuid)
        return res

    def get_usage_imp(self, uuid):
        xray_client = self.get_xray_client()
        d = xray_client.get_client_download_traffic(f'{uuid}@hiddify.com', reset=True)
        u = xray_client.get_client_upload_traffic(f'{uuid}@hiddify.com', reset=True)

        res = None
        if d is None:
            res = u
        elif u is None:
            res = d
        else:
            res = d + u
        if res:
            print(f"Xray usage {uuid} d={d} u={u} sum={res}")
        return res



import datetime
import json
import os
import random
import sys
import uuid


from hiddifypanel import Events, hutils
from hiddifypanel.cache import cache
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
from hiddifypanel.database import db, db_execute
from flask import g
from sqlalchemy import text

MAX_DB_VERSION = 90


def _v84(child_id):
    # the 2022-blake3-chacha20-poly1305 encryption method doesn't support multiuser config
    if hconfig(ConfigEnum.shadowsocks2022_method) == '2022-blake3-chacha20-poly1305':
        set_hconfig(ConfigEnum.shadowsocks2022_method, '2022-blake3-aes-256-gcm')


def _v83(child_id):
    set_hconfig(ConfigEnum.log_level, LogLevel.CRITICAL)


def _v82(child_id):
    set_hconfig(ConfigEnum.vless_enable, True)
    set_hconfig(ConfigEnum.trojan_enable, True)
    set_hconfig(ConfigEnum.reality_enable, True)
    set_hconfig(ConfigEnum.tcp_enable, True)
    set_hconfig(ConfigEnum.quic_enable, True)
    set_hconfig(ConfigEnum.xtls_enable, True)
    set_hconfig(ConfigEnum.h2_enable, True)


def _v80(child_id):
    set_hconfig(ConfigEnum.parent_domain, '')
    set_hconfig(ConfigEnum.parent_admin_proxy_path, '')


def _v79(child_id):
    set_hconfig(ConfigEnum.panel_mode, PanelMode.standalone)


def _v78(child_id):
    # equalize panel unique id and root child unique id
    root_child_unique_id = Child.query.filter(Child.name == "Root").first().unique_id
    set_hconfig(ConfigEnum.unique_id, root_child_unique_id)


def _v77(child_id):
    pass


def _v75(child_id):
    for u in User.query.all():
        hutils.model.gen_wg_keys(u)


def _v74(child_id):
    set_hconfig(ConfigEnum.ws_enable, True)
    set_hconfig(ConfigEnum.grpc_enable, True)
    set_hconfig(ConfigEnum.httpupgrade_enable, True)
    set_hconfig(ConfigEnum.shadowsocks2022_port, hutils.random.get_random_unused_port())
    set_hconfig(ConfigEnum.shadowsocks2022_method, "2022-blake3-aes-256-gcm")
    set_hconfig(ConfigEnum.shadowsocks2022_enable, False)
    set_hconfig(ConfigEnum.path_httpupgrade, hutils.random.get_random_string(7, 15))
    db.session.bulk_save_objects(get_proxy_rows_v1())

    for i in range(1, 10):
        for d in hutils.network.get_random_domains(50):
            if hutils.network.is_domain_reality_friendly(d):
                set_hconfig(ConfigEnum.shadowtls_fakedomain, d)
                return
    set_hconfig(ConfigEnum.shadowtls_fakedomain, "captive.apple.com")


def _v71(child_id):
    add_config_if_not_exist(ConfigEnum.tuic_port, hutils.random.get_random_unused_port())
    add_config_if_not_exist(ConfigEnum.hysteria_port, hutils.random.get_random_unused_port())
    add_config_if_not_exist(ConfigEnum.ssh_server_port, hutils.random.get_random_unused_port())
    add_config_if_not_exist(ConfigEnum.wireguard_port, hutils.random.get_random_unused_port())


def _v70(child_id):
    Domain.query.filter(Domain.child_id != 0).delete()
    StrConfig.query.filter(StrConfig.child_id != 0).delete()
    BoolConfig.query.filter(BoolConfig.child_id != 0).delete()
    Proxy.query.filter(Proxy.child_id != 0).delete()
    Child.query.filter(Child.id != 0).delete()

    child = Child.by_id(0)
    child.unique_id = str(uuid.uuid4())
    child.type = ChildMode.virtual


# using child_id in lower version is not needed as it is introduced in v70


def _v69():
    db.session.bulk_save_objects(get_proxy_rows_v1())
    add_config_if_not_exist(ConfigEnum.wireguard_enable, True)
    add_config_if_not_exist(ConfigEnum.wireguard_port, hutils.random.get_random_unused_port())
    add_config_if_not_exist(ConfigEnum.wireguard_ipv4, "10.90.0.1")
    add_config_if_not_exist(ConfigEnum.wireguard_ipv6, "fd42:42:90::1")
    wg_pk, wg_pub, _ = hutils.crypto.get_wg_private_public_psk_pair()
    add_config_if_not_exist(ConfigEnum.wireguard_private_key, wg_pk)
    add_config_if_not_exist(ConfigEnum.wireguard_public_key, wg_pub)
    add_config_if_not_exist(ConfigEnum.wireguard_noise_trick, "5-10")
    for u in User.query.all():
        u.wg_pk, u.wg_pub, u.wg_psk = hutils.crypto.get_wg_private_public_psk_pair()


def _v65():
    add_config_if_not_exist(ConfigEnum.mux_enable, False)
    add_config_if_not_exist(ConfigEnum.mux_protocol, 'smux')
    add_config_if_not_exist(ConfigEnum.mux_max_connections, '4')
    add_config_if_not_exist(ConfigEnum.mux_min_streams, '4')
    add_config_if_not_exist(ConfigEnum.mux_max_streams, '0')
    add_config_if_not_exist(ConfigEnum.mux_padding_enable, False)
    add_config_if_not_exist(ConfigEnum.mux_brutal_enable, True)
    add_config_if_not_exist(ConfigEnum.mux_brutal_up_mbps, '100')
    add_config_if_not_exist(ConfigEnum.mux_brutal_down_mbps, '100')


def _v63():
    add_config_if_not_exist(ConfigEnum.hysteria_enable, True)
    add_config_if_not_exist(ConfigEnum.hysteria_port, hutils.random.get_random_unused_port())
    add_config_if_not_exist(ConfigEnum.hysteria_obfs_enable, True)
    add_config_if_not_exist(ConfigEnum.hysteria_up_mbps, "150")
    add_config_if_not_exist(ConfigEnum.hysteria_down_mbps, "300")


def _v62():
    add_config_if_not_exist(ConfigEnum.tls_fragment_enable, False)
    add_config_if_not_exist(ConfigEnum.tls_fragment_size, "10-100")
    add_config_if_not_exist(ConfigEnum.tls_fragment_sleep, "50-200")
    add_config_if_not_exist(ConfigEnum.tls_mixed_case, False)
    add_config_if_not_exist(ConfigEnum.tls_padding_enable, False)
    add_config_if_not_exist(ConfigEnum.tls_padding_length, "50-200")


def _v61():
    execute("ALTER TABLE user MODIFY COLUMN username VARCHAR(100);")
    execute("ALTER TABLE user MODIFY COLUMN password VARCHAR(100);")


def _v60():
    add_config_if_not_exist(ConfigEnum.proxy_path_admin, hutils.random.get_random_string())
    add_config_if_not_exist(ConfigEnum.proxy_path_client, hutils.random.get_random_string())


def _v59():
    # set user model username and password
    for u in User.query.all():
        hutils.model.gen_username(u)
        hutils.model.gen_password(u)

    # set admin model username and password
    for a in AdminUser.query.all():
        hutils.model.gen_username(a)
        hutils.model.gen_password(a)


def _v57():
    add_config_if_not_exist(ConfigEnum.warp_sites, "")


def _v56():
    set_hconfig(ConfigEnum.reality_port, hutils.random.get_random_unused_port())


def _v55():
    tuic_port = hutils.random.get_random_unused_port()
    hystria_port = hutils.random.get_random_unused_port()
    set_hconfig(ConfigEnum.tuic_port, tuic_port)
    set_hconfig(ConfigEnum.hysteria_port, hystria_port)
    set_hconfig(ConfigEnum.tuic_enable, True)
    set_hconfig(ConfigEnum.hysteria_enable, True)
    Proxy.query.filter(Proxy.proto.in_(["tuic", "hysteria2", "hysteria"])).delete()
    db.session.add(Proxy(l3='tls', transport='custom', cdn='direct', proto='tuic', enable=True, name="TUIC"))
    db.session.add(Proxy(l3='tls', transport='custom', cdn='direct', proto='hysteria2', enable=True, name="Hysteria2"))


def _v52():
    db.session.bulk_save_objects(get_proxy_rows_v1())


def _v51():
    Proxy.query.filter(Proxy.l3.in_([ProxyL3.h3_quic])).delete()


def _v50():
    set_hconfig(ConfigEnum.show_usage_in_sublink, True)


def _v49():

    for u in User.query.all():
        priv, publ = hutils.crypto.get_ed25519_private_public_pair()
        u.ed25519_private_key = priv
        u.ed25519_public_key = publ


def _v48():
    add_config_if_not_exist(ConfigEnum.ssh_server_enable, True)
    set_hconfig(ConfigEnum.ssh_server_enable, True)


def _v47():
    StrConfig.query.filter(StrConfig.key == ConfigEnum.ssh_server_enable).delete()


def _v45():

    if not Proxy.query.filter(Proxy.name == "SSH").first():
        db.session.add(Proxy(l3='ssh', transport='ssh', cdn='direct', proto='ssh', enable=True, name="SSH"))

    add_config_if_not_exist(ConfigEnum.ssh_server_port, hutils.random.get_random_unused_port())
    add_config_if_not_exist(ConfigEnum.ssh_server_enable, False)
# def _v43():
#     if not (Domain.query.filter(Domain.domain==hconfig(ConfigEnum.domain_fronting_domain)).first()):
#         db.session.add(Domain(domain=hconfig(ConfigEnum.domain_fronting_domain),servernames=hconfig(ConfigEnum.domain_fronting_domain),mode=DomainType.cdn))

# v7.0.0


def _v42():

    for k in [ConfigEnum.telegram_fakedomain, ConfigEnum.ssfaketls_fakedomain, ConfigEnum.shadowtls_fakedomain]:
        if not hconfig(k):
            rnd_domains = hutils.network.get_random_domains(1)
            add_config_if_not_exist(k, rnd_domains[0])


def _v41():
    add_config_if_not_exist(ConfigEnum.core_type, "xray")
    if not (Domain.query.filter(Domain.domain == hconfig(ConfigEnum.reality_fallback_domain)).first()):
        db.session.add(Domain(domain=hconfig(ConfigEnum.reality_fallback_domain),
                       servernames=hconfig(ConfigEnum.reality_server_names), mode=DomainType.reality))


def _v38():
    add_config_if_not_exist(ConfigEnum.dns_server, "1.1.1.1")
    add_config_if_not_exist(ConfigEnum.warp_mode, "all" if hconfig(ConfigEnum.warp_enable) else "none")
    add_config_if_not_exist(ConfigEnum.warp_plus_code, '')


# def _v34():
#     add_config_if_not_exist(ConfigEnum.show_usage_in_sublink, True)


def _v33():
    Proxy.query.filter(Proxy.l3 == ProxyL3.reality).delete()
    _v31()


def _v31():
    add_config_if_not_exist(ConfigEnum.reality_short_ids, uuid.uuid4().hex[0:random.randint(1, 8) * 2])
    key_pair = hutils.crypto.generate_x25519_keys()
    add_config_if_not_exist(ConfigEnum.reality_private_key, key_pair['private_key'])
    add_config_if_not_exist(ConfigEnum.reality_public_key, key_pair['public_key'])
    db.session.bulk_save_objects(get_proxy_rows_v1())
    if not (AdminUser.query.filter(AdminUser.id == 1).first()):
        db.session.add(AdminUser(id=1, uuid=hconfig(ConfigEnum.admin_secret), name="Owner", mode=AdminMode.super_admin, comment=""))
        execute("update admin_user set id=1 where name='owner'")
    for i in range(1, 10):
        for d in hutils.network.get_random_domains(50):
            if hutils.network.is_domain_reality_friendly(d):
                add_config_if_not_exist(ConfigEnum.reality_fallback_domain, d)
                add_config_if_not_exist(ConfigEnum.reality_server_names, d)
                return
    add_config_if_not_exist(ConfigEnum.reality_fallback_domain, "yahoo.com")
    add_config_if_not_exist(ConfigEnum.reality_server_names, "yahoo.com")

    # add_config_if_not_exist(ConfigEnum.cloudflare, "")


def _v27():
    # add_config_if_not_exist(ConfigEnum.cloudflare, "")
    set_hconfig(ConfigEnum.netdata, False)


def _v26():
    add_config_if_not_exist(ConfigEnum.cloudflare, "")
    add_config_if_not_exist(ConfigEnum.country, "ir")
    add_config_if_not_exist(ConfigEnum.parent_panel, "")
    add_config_if_not_exist(ConfigEnum.is_parent, False)
    add_config_if_not_exist(ConfigEnum.license, "")


def _v21():
    db.session.bulk_save_objects(get_proxy_rows_v1())


def _v20():
    if hconfig(ConfigEnum.domain_fronting_domain):
        fake_domains = [hconfig(ConfigEnum.domain_fronting_domain)]

        direct_domain = Domain.query.filter(Domain.mode in [DomainType.direct, DomainType.relay]).first()
        if direct_domain:
            direct_host = direct_domain.domain
        else:
            direct_host = hutils.network.get_ip_str(4)

        for fd in fake_domains:
            if not Domain.query.filter(Domain.domain == fd).first():
                db.session.add(Domain(domain=fd, mode='fake', alias='moved from domain fronting', cdn_ip=direct_host))


def _v19():
    set_hconfig(ConfigEnum.path_trojan, hutils.random.get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_vless, hutils.random.get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_vmess, hutils.random.get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_ss, hutils.random.get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_grpc, hutils.random.get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_tcp, hutils.random.get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_ws, hutils.random.get_random_string(7, 15))
    add_config_if_not_exist(ConfigEnum.tuic_enable, False)
    add_config_if_not_exist(ConfigEnum.shadowtls_enable, False)
    add_config_if_not_exist(ConfigEnum.shadowtls_fakedomain, "en.wikipedia.org")
    add_config_if_not_exist(ConfigEnum.utls, "chrome")
    add_config_if_not_exist(ConfigEnum.telegram_bot_token, "")
    add_config_if_not_exist(ConfigEnum.package_mode, "release")


def _v17():
    for u in User.query.all():
        if u.expiry_time:
            if not u.package_days:
                if not u.last_reset_time:
                    u.package_days = (u.expiry_time - datetime.date.today()).days
                    u.start_date = datetime.date.today()
                else:
                    u.package_days = (u.expiry_time - u.last_reset_time).days
                    u.start_date = u.last_reset_time
            u.expiry_time = None


def _v1():
    external_ip = str(hutils.network.get_ip_str(4))
    rnd_domains = hutils.network.get_random_domains(5)

    data = [
        StrConfig(key=ConfigEnum.db_version, value=1),
        User(name="default", usage_limit_GB=3000, package_days=3650, mode=UserMode.weekly),
        Domain(domain=external_ip + ".sslip.io", mode=DomainType.direct),
        StrConfig(key=ConfigEnum.admin_secret, value=uuid.uuid4()),
        StrConfig(key=ConfigEnum.http_ports, value="80"),
        StrConfig(key=ConfigEnum.tls_ports, value="443"),
        BoolConfig(key=ConfigEnum.first_setup, value=True),
        StrConfig(key=ConfigEnum.decoy_domain, value=hutils.network.get_random_decoy_domain()),
        StrConfig(key=ConfigEnum.proxy_path, value=hutils.random.get_random_string()),
        BoolConfig(key=ConfigEnum.firewall, value=False),
        BoolConfig(key=ConfigEnum.netdata, value=True),
        StrConfig(key=ConfigEnum.lang, value='en'),
        BoolConfig(key=ConfigEnum.block_iran_sites, value=True),
        BoolConfig(key=ConfigEnum.allow_invalid_sni, value=True),
        BoolConfig(key=ConfigEnum.kcp_enable, value=False),
        StrConfig(key=ConfigEnum.kcp_ports, value="88"),
        BoolConfig(key=ConfigEnum.auto_update, value=True),
        BoolConfig(key=ConfigEnum.speed_test, value=True),
        BoolConfig(key=ConfigEnum.only_ipv4, value=False),
        BoolConfig(key=ConfigEnum.vmess_enable, value=True),
        BoolConfig(key=ConfigEnum.http_proxy_enable, value=True),
        StrConfig(key=ConfigEnum.shared_secret, value=str(uuid.uuid4())),
        BoolConfig(key=ConfigEnum.telegram_enable, value=True),
        # StrConfig(key=ConfigEnum.telegram_secret,value=uuid.uuid4().hex),
        StrConfig(key=ConfigEnum.telegram_adtag, value=""),
        StrConfig(key=ConfigEnum.telegram_fakedomain, value=rnd_domains[1]),
        BoolConfig(key=ConfigEnum.ssfaketls_enable, value=False),
        # StrConfig(key=ConfigEnum.ssfaketls_secret,value=str(uuid.uuid4())),
        StrConfig(key=ConfigEnum.ssfaketls_fakedomain, value=rnd_domains[2]),
        BoolConfig(key=ConfigEnum.shadowtls_enable, value=False),
        # StrConfig(key=ConfigEnum.shadowtls_secret,value=str(uuid.uuid4())),
        StrConfig(key=ConfigEnum.shadowtls_fakedomain, value=rnd_domains[3]),

        BoolConfig(key=ConfigEnum.ssr_enable, value=False),
        # StrConfig(key=ConfigEnum.ssr_secret,value=str(uuid.uuid4())),
        StrConfig(key=ConfigEnum.ssr_fakedomain, value=rnd_domains[4]),

        # BoolConfig(key=ConfigEnum.tuic_enable, value=False),
        # StrConfig(key=ConfigEnum.tuic_port, value=3048),

        BoolConfig(key=ConfigEnum.domain_fronting_tls_enable, value=False),
        BoolConfig(key=ConfigEnum.domain_fronting_http_enable, value=False),
        StrConfig(key=ConfigEnum.domain_fronting_domain, value=""),

        # BoolConfig(key=ConfigEnum.torrent_block,value=False),

        *get_proxy_rows_v1()
    ]
    # fake_domains=['speedtest.net']
    # for fd in fake_domains:
    #     if not Domain.query.filter(Domain.domain==fd).first():
    #         db.session.add(Domain(domain=fd,mode='fake',alias='fake domain',cdn_ip=external_ip))
    db.session.bulk_save_objects(data)


def _v7():
    try:
        Proxy.query.filter(Proxy.name == 'tls XTLS direct trojan').delete()
        Proxy.query.filter(Proxy.name == 'tls XTLSVision direct trojan').delete()
    except BaseException:
        pass
    add_config_if_not_exist(ConfigEnum.telegram_lib, "erlang")
    add_config_if_not_exist(ConfigEnum.admin_lang, hconfig(ConfigEnum.lang))
    add_config_if_not_exist(ConfigEnum.branding_title, "")
    add_config_if_not_exist(ConfigEnum.branding_site, "")
    add_config_if_not_exist(ConfigEnum.branding_freetext, "")
    add_config_if_not_exist(ConfigEnum.v2ray_enable, False)
    add_config_if_not_exist(ConfigEnum.is_parent, False)
    add_config_if_not_exist(ConfigEnum.parent_panel, '')
    add_config_if_not_exist(ConfigEnum.unique_id, str(uuid.uuid4()))


def _v9():
    # add_column(User.mode)
    # add_column(User.comment)
    try:
        for u in User.query.all():
            u.mode = UserMode.monthly if u.monthly else UserMode.no_reset
    except BaseException:
        pass


def _v10():
    all_configs = get_hconfigs()
    execute("ALTER TABLE `str_config` RENAME TO `str_config_old`")
    execute("ALTER TABLE `bool_config` RENAME TO `bool_config_old`")
    # db.create_all()
    rows = []
    for c, v in all_configs.items():
        if c.type == bool:
            rows.append(BoolConfig(key=c, value=v, child_id=0))
        else:
            rows.append(StrConfig(key=c, value=v, child_id=0))

    db.session.bulk_save_objects(rows)


def get_proxy_rows_v1():
    rows = list(make_proxy_rows([
        "h2 direct vless",
        "XTLS direct vless",
        "WS direct vless",
        "WS direct trojan",
        "WS direct vmess",
        "httpupgrade direct vless",
        # "httpupgrade direct trojan",
        "httpupgrade direct vmess",
        "tcp direct vless",
        "tcp direct trojan",
        "tcp direct vmess",
        "grpc direct vless",
        "grpc direct trojan",
        "grpc direct vmess",
        "faketls direct ss",
        "WS direct v2ray",
        "h2 relay vless",
        "XTLS relay vless",
        "WS relay vless",
        "WS relay trojan",
        "WS relay vmess",
        "httpupgrade relay vless",
        # "httpupgrade relay trojan",
        "httpupgrade relay vmess",
        "tcp relay vless",
        "tcp relay trojan",
        "tcp relay vmess",
        "grpc relay vless",
        "grpc relay trojan",
        "grpc relay vmess",
        "faketls relay ss",
        "WS relay v2ray",

        # "restls1_2 direct ss",
        # "restls1_3 direct ss",
        # "tcp direct ssr",
        "WS CDN v2ray",
        "WS CDN vless",
        "WS CDN trojan",
        "WS CDN vmess",
        "httpupgrade CDN vless",
        # "httpupgrade CDN trojan",
        "httpupgrade CDN vmess",
        "grpc CDN vless",
        "grpc CDN trojan",
        "grpc CDN vmess",

    ]
    ))
    rows.append(Proxy(l3=ProxyL3.custom, transport=ProxyTransport.shadowsocks, cdn='direct', proto='ss', enable=True, name="ShadowSocks2022"))
    rows.append(Proxy(l3=ProxyL3.custom, transport=ProxyTransport.shadowsocks, cdn='relay', proto='ss', enable=True, name="ShadowSocks2022 Relay"))

    rows.append(Proxy(l3=ProxyL3.tls, transport=ProxyTransport.shadowtls, cdn='direct', proto='ss', enable=True, name="ShadowTLS"))
    rows.append(Proxy(l3=ProxyL3.tls, transport=ProxyTransport.shadowtls, cdn='relay', proto='ss', enable=True, name="ShadowTLS Relay"))
    rows.append(Proxy(l3='ssh', transport='ssh', cdn='direct', proto='ssh', enable=True, name="SSH"))
    rows.append(Proxy(l3='ssh', transport=ProxyTransport.ssh, cdn=ProxyCDN.relay, proto=ProxyProto.ssh, enable=True, name="SSH Relay"))

    rows.append(Proxy(l3='tls', transport='custom', cdn='direct', proto='tuic', enable=True, name="TUIC"))
    rows.append(Proxy(l3='tls', transport='custom', cdn='relay', proto='tuic', enable=True, name="TUIC Relay"))
    rows.append(Proxy(l3='tls', transport='custom', cdn='direct', proto='hysteria2', enable=True, name="Hysteria2"))
    rows.append(Proxy(l3='tls', transport='custom', cdn='relay', proto='hysteria2', enable=True, name="Hysteria2 Relay"))
    rows.append(Proxy(l3=ProxyL3.udp, transport=ProxyTransport.custom, cdn=ProxyCDN.direct, proto=ProxyProto.wireguard, enable=True, name="WireGuard"))
    rows.append(Proxy(l3=ProxyL3.udp, transport=ProxyTransport.custom, cdn=ProxyCDN.relay, proto=ProxyProto.wireguard, enable=True, name="WireGuard Relay"))
    for p in rows:
        is_exist = Proxy.query.filter(Proxy.name == p.name).first() or Proxy.query.filter(
            Proxy.l3 == p.l3, Proxy.transport == p.transport, Proxy.cdn == p.cdn, Proxy.proto == p.proto).first()
        if not is_exist:
            yield p


def make_proxy_rows(cfgs):
    # "h3_quic",
    for l3 in ["tls_h2", "tls", "http", "reality"]:
        for c in cfgs:
            transport, cdn, proto = c.split(" ")
            if l3 in ["kcp", 'reality'] and cdn != "direct":
                continue
            if l3 == "reality" and ((transport not in ['tcp', 'grpc', 'XTLS']) or proto != 'vless'):
                continue
            if proto == "trojan" and l3 not in ["tls", 'xtls', 'tls_h2', 'h3_quic']:
                continue
            if transport in ["grpc", "XTLS", "faketls"] and l3 == "http":
                continue
            if transport in ["h2"] and l3 != "reality":
                continue
            # if l3 == "tls_h2" and transport =="grpc":
            #     continue
            enable = l3 != "http" or proto == "vmess"
            enable = enable and transport != 'tcp'
            name = f'{l3} {c}'
            # is_exist = Proxy.query.filter(Proxy.name == name).first() or Proxy.query.filter(
            #     Proxy.l3 == l3, Proxy.transport == transport, Proxy.cdn == cdn, Proxy.proto == proto).first()
            # if not is_exist:
            yield Proxy(l3=l3, transport=transport, cdn=cdn, proto=proto, enable=enable, name=name)


def add_config_if_not_exist(key: ConfigEnum, val: str | int, child_id: int | None = None):
    if child_id is None:
        child_id = Child.current().id

    old_val = hconfig(key, child_id)
    hutils.utils.error(f'{key}, {val}, {child_id}, {old_val}')
    if old_val is None:
        set_hconfig(key, val)


def add_column(column):
    try:
        column_type = column.type.compile(db.engine.dialect)

        db_execute(f'ALTER TABLE {column.table.name} ADD COLUMN {column.name} {column_type}', commit=True)
    except BaseException:
        pass


def execute(query: str):
    try:

        db_execute(query)
    except BaseException as e:
        hutils.utils.error(e)
        pass


def add_new_enum_values():
    columns = [
        Proxy.l3, Proxy.proto, Proxy.cdn, Proxy.transport,
        User.mode, Domain.mode, BoolConfig.key, StrConfig.key
    ]
    for col in columns:
        enum_class = col.type.enum_class
        column_name = col.name
        table_name = col.table

        # Get the existing values in the enum
        existing_values = [f'{e}' if isinstance(e, ConfigEnum) else e.value for e in enum_class]
        # print("existing_values--------------", existing_values)
        # Get the values in the enum column in the database
        # result = db.engine.execute(f"SELECT DISTINCT `{column_name}` FROM {table_name}")
        # db_values = {row[0] for row in result}

        result = db.session.execute(text(f"SHOW COLUMNS FROM {table_name} LIKE '{column_name}';")).fetchall()
        db_values = []

        for row in result:
            if "enum" in row[1]:
                db_values = row[1][5:-1].split(",")
                break
        db_values = [value.strip("'") for value in db_values]

        # Find the new values that need to be added to the enum column in the database
        new_values = set(existing_values) - set(db_values)
        old_values = set(db_values) - set(existing_values)
        # print('new_values-=-----------------', new_values)
        if len(new_values) == 0 and len(old_values) == 0:
            continue

        # Add the new value to the enum column in the database
        enumstr = ','.join([f"'{a}'" for a in [*existing_values, *old_values]])

        db_execute(f"ALTER TABLE {table_name} MODIFY COLUMN `{column_name}` ENUM({enumstr});", commit=True)


def latest_db_version():
    for ver in range(MAX_DB_VERSION, 1, -1):
        db_action = sys.modules[__name__].__dict__.get(f'_v{ver}', None)
        if db_action:
            return ver
    return 0


def upgrade_database():

    panel_root = '/opt/hiddify-manager/hiddify-panel/'
    backup_root = f"{panel_root}backup/"
    sqlite_db = f"{panel_root}hiddifypanel.db"
    if not os.path.isdir(backup_root) or len(os.listdir(backup_root)) == 0:
        if os.path.isfile(sqlite_db):
            os.rename(sqlite_db, sqlite_db + ".old")
        hutils.utils.error("no backup found...")
        return
    if os.path.isfile(sqlite_db):
        hutils.utils.error("Finding Old Version Database... importing configs from latest backup")
        newest_file = max([(f, os.path.getmtime(os.path.join(backup_root, f)))
                          for f in os.listdir(backup_root) if os.path.isfile(os.path.join(backup_root, f))], key=lambda x: x[1])[0]
        with open(f'{backup_root}{newest_file}', 'r') as f:
            hutils.utils.error(f"importing configs from {newest_file}")
            json_data = json.load(f)
            hiddify.set_db_from_json(json_data,
                                     set_users=True,
                                     set_domains=True,
                                     remove_domains=True,
                                     remove_users=True,
                                     set_settings=True,
                                     override_unique_id=True,
                                     set_admins=True,
                                     override_root_admin=True,
                                     override_child_unique_id=0,
                                     replace_owner_admin=True
                                     )
            db_version = int([d['value'] for d in json_data['hconfigs'] if d['key'] == "db_version"][0])
            os.rename(sqlite_db, sqlite_db + ".old")
            set_hconfig(ConfigEnum.db_version, db_version, commit=True)

        hutils.utils.error("Upgrading to the new dataset succuess.")


def init_db():
    db.create_all()
    cache.invalidate_all_cached_functions()

    # set_hconfig(ConfigEnum.db_version, 71)
    # temporary fix
    add_column(Child.mode)
    add_column(Child.name)
    db_version = int(hconfig(ConfigEnum.db_version) or 0)
    if db_version == latest_db_version():
        return
    migrate(db_version)
    Child.query.filter(Child.id == 0).first().mode = ChildMode.virtual
    # if db_version < 69:
    #     _v70(0)

    db.session.commit()

    for child in Child.query.filter(Child.mode == ChildMode.virtual).all():
        g.child = child
        db_version = int(hconfig(ConfigEnum.db_version, child.id) or 0)
        start_version = db_version
        for ver in range(1, MAX_DB_VERSION):
            if ver <= db_version:
                continue

            db_action = sys.modules[__name__].__dict__.get(f'_v{ver}', None)
            if not db_action or (start_version == 0 and ver == 10):
                continue

            hutils.utils.error(f"Updating db from version {db_version} for node {child.id}")

            if ver < 70:
                if child.id != 0:
                    continue
                db_action()
            else:
                db_action(child.id)

            Events.db_init_event.notify(db_version=db_version)
            hutils.utils.error(f"Updated successfuly db from version {db_version} to {ver}")

            db_version = ver
            db.session.commit()
            set_hconfig(ConfigEnum.db_version, db_version, child_id=child.id, commit=False)

        db.session.commit()
    g.child = Child.by_id(0)
    return BoolConfig.query.all()


def migrate(db_version):
    for table_name, table_obj in db.metadata.tables.items():
        for column in table_obj.columns:
            add_column(column)
    Events.db_prehook.notify()
    if db_version < 82:
        execute('ALTER TABLE child DROP INDEX `name`;')
    if db_version < 77:
        execute('ALTER TABLE user_detail DROP COLUMN connected_ips;')
        execute('update user_detail set connected_devices="" where connected_devices IS NULL')

    if db_version < 70:
        execute('CREATE INDEX date ON daily_usage (date);')
        execute('CREATE INDEX username ON user (username);')
        execute('CREATE INDEX username ON admin_user (username);')
        execute('CREATE INDEX telegram_id ON user (telegram_id);')
        execute('CREATE INDEX telegram_id ON admin_user (telegram_id);')

        execute('ALTER TABLE proxy DROP INDEX `name`;')

        execute("ALTER TABLE user MODIFY COLUMN telegram_id BIGINT;")
        execute("ALTER TABLE admin_user MODIFY COLUMN telegram_id BIGINT;")

        # aaa
        # # add_column(UserDetail.connected_devices)
        # add_column(Child.mode)
        # add_column(Child.name)
        # add_column(User.lang)
        # add_column(AdminUser.lang)
        # add_column(User.username)
        # add_column(User.password)
        # add_column(AdminUser.username)
        # add_column(AdminUser.password)
        # add_column(User.wg_pk)
        # add_column(User.wg_pub)
        # add_column(User.wg_psk)

        # add_column(Domain.extra_params)

    if db_version < 52:
        execute(f'update domain set mode="sub_link_only", sub_link_only=false where sub_link_only = true or mode=1  or mode="1"')
        execute(f'update domain set mode="direct", sub_link_only=false where mode=0  or mode="0"')
        execute(f'update proxy set transport="WS" where transport = "ws"')
        execute(f'update admin_user set mode="agent" where mode = "slave"')
        execute(f'update admin_user set mode="super_admin" where id=1')
        execute(f'DELETE from proxy where transport = "h1"')
        # add_column(Domain.grpc)
        # add_column(ParentDomain.alias)
        # add_column(User.ed25519_private_key)
        # add_column(User.ed25519_public_key)
        # add_column(User.start_date)
        # add_column(User.package_days)
        # add_column(User.telegram_id)
        # add_column(Child.unique_id)
        # add_column(Domain.alias)
        # add_column(Domain.sub_link_only)
        # add_column(Domain.child_id)
        # add_column(Proxy.child_id)
        # add_column(User.added_by)
        # add_column(User.max_ips)
        # add_column(AdminUser.parent_admin_id)
        # add_column(AdminUser.can_add_admin)
        # add_column(AdminUser.max_active_users)
        # add_column(AdminUser.max_users)
        # add_column(BoolConfig.child_id)
        # add_column(StrConfig.child_id)
        # add_column(DailyUsage.admin_id)
        # add_column(DailyUsage.child_id)
        # add_column(User.monthly)
        # add_column(User.enable)
        # add_column(Domain.cdn_ip)
        # add_column(Domain.servernames)
        # add_column(User.lang)

        if len(Domain.query.all()) != 0 and BoolConfig.query.count() == 0:
            execute(f'DROP TABLE bool_config')
            execute(f'ALTER TABLE bool_config_old RENAME TO bool_config')
        if len(Domain.query.all()) != 0 and StrConfig.query.count() == 0:
            execute(f'DROP TABLE str_config')
            execute(f'ALTER TABLE str_config_old RENAME TO str_config')

        execute('ALTER TABLE user RENAME COLUMN monthly_usage_limit_GB TO usage_limit_GB')
        execute(f'update admin_user set parent_admin_id=1 where parent_admin_id is NULL and 1!=id')
        execute(f'update admin_user set max_users=100,max_active_users=100 where max_users is NULL')
        execute(f'update dailyusage set child_id=0 where child_id is NULL')
        execute(f'update dailyusage set admin_id=1 where admin_id is NULL')
        execute(f'update dailyusage set admin_id=1 where admin_id = 0')
        execute(f'update user set added_by=1 where added_by = 1')
        execute(f'update user set enable=True, mode="no_reset" where enable is NULL')
        execute(f'update user set enable=False, mode="no_reset" where mode = "disable"')
        execute(f'update user set added_by=1 where added_by is NULL')
        execute(f'update user set max_ips=10000 where max_ips is NULL')
        execute(f'update str_config set child_id=0 where child_id is NULL')
        execute(f'update bool_config set child_id=0 where child_id is NULL')
        execute(f'update domain set child_id=0 where child_id is NULL')
        execute(f'update domain set sub_link_only=False where sub_link_only is NULL')
        execute(f'update proxy set child_id=0 where child_id is NULL')

    add_new_enum_values()

    AdminUser.get_super_admin()  # to create super admin if not exist

    upgrade_database()
    db.session.commit()



from urllib.parse import urlparse
from flask.views import MethodView
from apiflask import Schema
from apiflask.fields import Integer, String, Float, URL, Enum, Boolean
from flask import g, request
from flask import current_app as app
from hiddifypanel import hutils, auth
from hiddifypanel.auth import login_required

from flask_babel import gettext as _
from hiddifypanel.models import Lang
from hiddifypanel.models.role import Role
from hiddifypanel.models.user import User
from hiddifypanel.models.config import hconfig
from hiddifypanel.models.config_enum import ConfigEnum
from hiddifypanel.database import db
from hiddifypanel.panel.user.user import get_common_data
from hiddifypanel.panel import hiddify


class ProfileSchema(Schema):
    profile_title = String(required=True)
    profile_url = URL(required=True)
    profile_usage_current = Float(required=True)
    profile_usage_total = Float(required=True)
    profile_remaining_days = Integer(required=True)
    profile_reset_days = Integer()
    telegram_bot_url = String()
    telegram_id = Integer()
    admin_message_html = String()
    admin_message_url = URL()
    brand_title = String()
    brand_icon_url = URL()
    doh = URL()
    lang = Enum(Lang, required=True)
    speedtest_enable = Boolean(required=True)
    telegram_proxy_enable = Boolean(required=True)


class UserInfoChangableSchema(Schema):
    language = Enum(Lang, required=False)
    telegram_id = Integer(required=False)


class InfoAPI(MethodView):
    decorators = [login_required({Role.user})]

    @app.output(ProfileSchema)
    def get(self):
        c = get_common_data(g.account.uuid, 'new')

        dto = ProfileSchema()
        # user is exist for sure
        dto.profile_title = c['user'].name
        dto.profile_url = f"{c['profile_url']}#{hutils.encode.url_encode(c['user'].name)}"
        dto.profile_usage_current = g.account.current_usage_GB
        dto.profile_usage_total = g.account.usage_limit_GB
        dto.profile_remaining_days = g.account.remaining_days
        dto.profile_reset_days = g.account.days_to_reset()
        dto.telegram_bot_url = f"https://t.me/{c['bot'].username}?start={g.account.uuid}" if c['bot'] else ""
        dto.telegram_id = c['user'].telegram_id

        dto.doh = f"https://{request.host}/{g.proxy_path}/dns/dns-query"
        dto.lang = (c['user'].lang) or Lang(hconfig(ConfigEnum.lang))
        dto.brand_icon_url = "" if hconfig(ConfigEnum.branding_title) else hutils.flask.static_url_for(filename="images/favicon.ico")
        # with force_locale("fa"):
        dto.admin_message_html = hconfig(ConfigEnum.branding_freetext) or _("Join our Hiddify Telegram channel to get the latest updates on Hiddify.")
        if not hconfig(ConfigEnum.branding_freetext) and auth.admin_session_is_exist():
            dto.admin_message_html += "<p style='font-style: italic;font-size:8px'>" + \
                _("[Admin only visible message:] You can change this message from settings") + "</p>"
        dto.admin_message_url = hconfig(ConfigEnum.branding_site) or "https://t.me/hiddify"
        dto.brand_title = hconfig(ConfigEnum.branding_title) or _("Hiddify")

        dto.speedtest_enable = hconfig(ConfigEnum.speed_test)
        dto.telegram_proxy_enable = c.get('telegram_enable', False)
        return dto

    @app.input(UserInfoChangableSchema, arg_name='data')
    def patch(self, data: UserInfoChangableSchema):
        if data['telegram_id'] and hutils.convert.is_int(data['telegram_id']):
            user = User.by_uuid(g.account.uuid)
            if user.telegram_id != data['telegram_id']:
                user.telegram_id = data['telegram_id']
                db.session.commit()

        if data['language']:
            user = User.by_uuid(g.account.uuid)
            if user.lang != data['language']:
                user.lang = data['language']
                db.session.commit()
        return {'message': 'ok'}



from flask_babel import lazy_gettext as _
from sqlalchemy import func
from typing import Dict
import datetime

from hiddifypanel.drivers import user_driver
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
from hiddifypanel.database import db
from hiddifypanel import hutils

to_gig_d = 1024**3


def update_local_usage():
    res = user_driver.get_users_usage(reset=True)
    return _add_users_usage(res, child_id=0)

    # return {"status": 'success', "comments":res}


def add_users_usage_uuid(uuids_bytes: Dict[str, Dict], child_id, sync=False):
    uuids_bytes = {u: v for u, v in uuids_bytes.items() if v}
    uuids = uuids_bytes.keys()
    users = User.query.filter(User.uuid.in_(uuids))
    dbusers_bytes = {u: uuids_bytes.get(u.uuid, 0) for u in users}
    _add_users_usage(dbusers_bytes, child_id, sync)  # type: ignore


def _add_users_usage(users_usage_data: Dict[User, Dict], child_id, sync=False):
    '''
    sync: when enabled, it means we have received usages from the parent panel
    '''
    res = {}
    have_change = False
    before_enabled_users = user_driver.get_enabled_users()
    daily_usage = {}
    today = datetime.date.today()
    for adm in AdminUser.query.all():
        daily_usage[adm.id] = DailyUsage.query.filter(DailyUsage.date == today, DailyUsage.admin_id == adm.id, DailyUsage.child_id == child_id).first()
        if not daily_usage[adm.id]:
            daily_usage[adm.id] = DailyUsage(admin_id=adm.id, child_id=child_id)
            db.session.add(daily_usage[adm.id])
        daily_usage[adm.id].online = User.query.filter(User.added_by == adm.id).filter(func.DATE(User.last_online) == today).count()
    # db.session.commit()
    userDetails = {p.user_id: p for p in UserDetail.query.filter(UserDetail.child_id == child_id).all()}
    for user, uinfo in users_usage_data.items():
        usage_bytes = uinfo['usage']
        devices = uinfo['devices']
        # user_active_before=user.is_active
        detail = userDetails.get(user.id)
        if not detail:
            detail = UserDetail(user_id=user.id, child_id=child_id)
            detail.current_usage_GB = detail.current_usage_GB or 0
            db.session.add(detail)
        detail.connected_devices = devices
        detail.current_usage_GB = detail.current_usage_GB or 0
        if not user.last_reset_time or user.user_should_reset():
            user.last_reset_time = datetime.date.today()
            user.current_usage_GB = 0
            detail.current_usage_GB = 0

        if not before_enabled_users[user.uuid] and user.is_active:
            print(f"Enabling disabled client {user.uuid} ")
            user_driver.add_client(user)
            send_bot_message(user)
            have_change = True
        if not isinstance(usage_bytes, int) or usage_bytes == 0:
            res[user.uuid] = "No usage"
        else:
            if sync:
                if daily_usage.get(user.added_by, daily_usage[1]).usage != usage_bytes:
                    daily_usage.get(user.added_by, daily_usage[1]).usage = usage_bytes
            else:
                daily_usage.get(user.added_by, daily_usage[1]).usage += usage_bytes
            in_gig = (usage_bytes) / to_gig_d
            res[user.uuid] = in_gig
            if sync:
                if user.current_usage_GB != in_gig:
                    user.current_usage_GB = in_gig
                # detail.current_usage_GB = in_gig
            else:
                user.current_usage_GB += in_gig
                detail.current_usage_GB += in_gig
            user.last_online = datetime.datetime.now()
            detail.last_online = datetime.datetime.now()

            if user.start_date is None:
                user.start_date = datetime.date.today()

        if before_enabled_users[user.uuid] and not user.is_active:
            print(f"Removing enabled client {user.uuid} ")
            user_driver.remove_client(user)
            have_change = True
            res[user.uuid] = f"{res[user.uuid]} !OUT of USAGE! Client Removed"

    db.session.commit()  # type: ignore
    if have_change:
        hiddify.quick_apply_users()

    # sync with the parent
    if not sync:
        if hutils.node.is_child():
            hutils.node.child.sync_users_usage_with_parent()
    return {"status": 'success', "comments": res}


def send_bot_message(user):
    if not (hconfig(ConfigEnum.telegram_bot_token) or not hutils.node.is_parent()):
        return
    if not user.telegram_id:
        return
    from hiddifypanel.panel.commercial.telegrambot import bot, Usage
    try:
        msg = Usage.get_usage_msg(user.uuid)
        msg = _("User activated!") if user.is_active else _("Package ended!") + "\n" + msg
        bot.send_message(user.telegram_id, msg, reply_markup=Usage.user_keyboard(user.uuid))
    except BaseException:
        pass




from urllib.parse import urlparse
from apiflask import Schema
from flask import g, request
from apiflask.fields import String, Integer
from flask import current_app as app
from flask.views import MethodView
from hiddifypanel.auth import login_required
from hiddifypanel.models.config import hconfig
from hiddifypanel.models.config_enum import ConfigEnum
from hiddifypanel.models.role import Role
from hiddifypanel.panel import hiddify


class ShortSchema(Schema):
    short = String(required=True, description="the short url slug")
    full_url = String(required=True, description="full short url")
    expire_in = Integer(required=True, description="expire_in is in seconds")


class ShortAPI(MethodView):
    decorators = [login_required({Role.user})]

    @app.output(ShortSchema)
    def get(self):
        short, expire_in = hiddify.add_short_link(hiddify.get_account_panel_link(g.account, request.host))
        full_url = f"https://{request.host}/{short}"
        dto = ShortSchema()
        dto.full_url = full_url
        dto.short = short
        # expire_in is in seconds
        dto.expire_in = expire_in
        return dto



from typing import List, Tuple
from flask import current_app, flash as flask_flash, g, request
from wtforms.validators import ValidationError
from apiflask import abort as apiflask_abort
from flask_babel import gettext as _
from flask import url_for  # type: ignore
from urllib.parse import urlparse
from markupsafe import Markup

import user_agents
import re
import os

from hiddifypanel.cache import cache
from hiddifypanel.models import *
from hiddifypanel import hutils


def flash(message: str, category: str = "message"):
    if not isinstance(message,str):
        message = str(message)
    return flask_flash(message, category)


def flash_config_success(restart_mode: ApplyMode = ApplyMode.nothing, domain_changed=True):
    if restart_mode != ApplyMode.nothing:
        url = hurl_for('admin.Actions:reinstall', complete_install=restart_mode == ApplyMode.restart, domain_changed=domain_changed)
        apply_btn = f"<a href='{url}' class='btn btn-primary form_post'>" + \
            _("admin.config.apply_configs") + "</a>"
        flash((_('config.validation-success', link=apply_btn)), 'success')  # type: ignore
    else:
        flash((_('config.validation-success-no-reset')), 'success')  # type: ignore


def static_url_for(**values):
    orig = url_for("static", **values)
    return orig.split("user_secret")[0]


def hurl_for(endpoint, **values):
    if Child.current().id != 0:

        new_endpoint = "child_" + endpoint
        if new_endpoint in current_app.view_functions:
            endpoint = new_endpoint
    return url_for(endpoint, **values)


def get_user_agent() -> dict:
    ua = __parse_user_agent(request.user_agent.string)

    if ua.get('v', 1) < 6:
        __parse_user_agent.invalidate_all()  # type:ignore
        ua = __parse_user_agent(request.user_agent.string)
    return ua


ua_version_pattern = re.compile(r'/(\d+\.\d+(\.\d+)?)')


@cache.cache()
def __parse_user_agent(ua: str) -> dict:
    # Example: SFA/1.8.0 (239; sing-box 1.8.0)
    # Example: SFA/1.7.0 (239; sing-box 1.7.0)
    # Example: HiddifyNext/0.13.6 (android) like ClashMeta v2ray sing-box

    uaa = user_agents.parse(request.user_agent.string)

    match = re.search(ua_version_pattern, request.user_agent.string)
    generic_version = list(map(int, match.group(1).split('.'))) if match else [0, 0, 0]
    res = {}
    res['v'] = 6
    res["is_bot"] = uaa.is_bot
    res["is_browser"] = re.match('^Mozilla', ua, re.IGNORECASE) and True
    res['os'] = uaa.os.family
    res['os_version'] = uaa.os.version
    res['is_clash'] = re.match('^(Clash|Stash)', ua, re.IGNORECASE) and True
    res['is_clash_meta'] = re.match('^(Clash-verge|Clash-?Meta|Stash|NekoBox|NekoRay|Pharos|hiddify-desktop)', ua, re.IGNORECASE) and True
    res['is_singbox'] = re.match('^(HiddifyNext|Dart|SFI|SFA)', ua, re.IGNORECASE) and True
    res['is_hiddify'] = re.match('^(HiddifyNext)', ua, re.IGNORECASE) and True
    res['is_streisand'] = re.match('^(Streisand)', ua, re.IGNORECASE) and True
    res['is_shadowrocket'] = re.match('^(Shadowrocket)', ua, re.IGNORECASE) and True
    res['is_v2rayng'] = re.match('^(v2rayNG)', ua, re.IGNORECASE) and True

    if res['is_singbox']:
        res['singbox_version'] = generic_version

    if ['is_hiddify']:
        res['hiddify_version'] = generic_version
        if generic_version[0] == 0 and generic_version[1] <= 14:
            res['singbox_version'] = [1, 7, 0]
        else:
            res['singbox_version'] = [1, 8, 0]

    res['is_v2ray'] = re.match('^(Hiddify|FoXray|Fair|v2rayNG|SagerNet|Shadowrocket|V2Box|Loon|Liberty)', ua, re.IGNORECASE) and True

    if res['os'] == 'Other':
        if re.match('^(FoXray|Fair|Shadowrocket|V2Box|Loon|Liberty)', ua, re.IGNORECASE):
            res['os'] = 'iOS'
            # res['os_version']

    for a in ['Hiddify', 'FoXray', 'Fair', 'v2rayNG', 'SagerNet', 'Shadowrocket', 'V2Box', 'Loon', 'Liberty', 'Clash', 'Meta', 'Stash', 'SFI', 'SFA', 'HiddifyNext']:
        if a.lower() in ua.lower():
            res['app'] = a
    if res["is_browser"]:
        res['app'] = uaa.browser.family
    return res


def get_proxy_path_from_url(url: str) -> str | None:
    url_path = urlparse(url).path
    proxy_path = url_path.lstrip('/').split('/')[0] or None
    return proxy_path


def is_api_call(req_path: str) -> bool:
    return 'api/v1/' in req_path or 'api/v2/' in req_path


def is_user_api_call() -> bool:
    if request.blueprint and request.blueprint == 'api_user':
        return True
    user_api_call_format = '/api/v2/user/'
    if user_api_call_format in request.path:
        return True
    return False


def is_user_panel_call(deprecated_format=False) -> bool:
    if request.blueprint and request.blueprint == 'client':
        return True
    if deprecated_format:
        user_panel_url = f'/{hconfig(ConfigEnum.proxy_path)}/'
    else:
        user_panel_url = f'/{hconfig(ConfigEnum.proxy_path_client)}/'
    if f'{request.path}'.startswith(user_panel_url) and "admin" not in f'{request.path}':
        return True
    return False


def is_admin_panel_call(deprecated_format: bool = False) -> bool:
    if request.blueprint and request.blueprint == 'admin':
        return True
    if deprecated_format:
        if f'{request.path}'.startswith(f'/{hconfig(ConfigEnum.proxy_path)}/') and "admin" in f'{request.path}':
            return True
    elif f'{request.path}'.startswith(f'/{hconfig(ConfigEnum.proxy_path_admin)}/admin/'):
        return True
    return False


def is_api_v1_call(endpoint=None) -> bool:
    if (request.blueprint and 'api_v1' in request.blueprint):
        return True
    elif endpoint and 'api_v1' in endpoint:
        return True
    elif request.endpoint and 'api_v1' in request.endpoint:
        return True

    api_v1_path = f'{request.host}/{hconfig(ConfigEnum.proxy_path_admin)}/api/v1/{AdminUser.get_super_admin_uuid()}/'
    if f'{request.host}{request.path}'.startswith(api_v1_path):
        return True
    return False


def is_login_call() -> bool:
    return request.blueprint == 'common_bp'


def is_admin_role(role: Role) -> bool:
    if not role:
        return False
    if role in {Role.super_admin, Role.admin, Role.agent}:
        return True
    return False


def is_admin_proxy_path() -> bool:
    proxy_path = g.get('proxy_path') or get_proxy_path_from_url(request.url)
    return proxy_path in [hconfig(ConfigEnum.proxy_path_admin)] or (proxy_path in [hconfig(ConfigEnum.proxy_path)] and "/admin/" in request.path)


def is_client_proxy_path() -> bool:
    proxy_path = g.get('proxy_path') or get_proxy_path_from_url(request.url)
    return proxy_path in [hconfig(ConfigEnum.proxy_path_client)] or (proxy_path in [hconfig(ConfigEnum.proxy_path)] and "/admin/" not in request.path)


def __is_admin_api_call() -> bool:
    if request.blueprint and request.blueprint == 'api_admin' or request.blueprint == 'api_v1':
        return True
    admin_api_call_format = '/api/v2/admin/'
    if admin_api_call_format in request.path:
        return True
    return False


def proxy_path_validator(proxy_path: str) -> None:
    # DEPRECATED PROXY_PATH HANDLED BY BACKWARD COMPATIBILITY MIDDLEWARE
    # does not nginx handle proxy path validation?

    if not proxy_path:
        return apiflask_abort(400, 'invalid request')

    dbg_mode = True if current_app.config['DEBUG'] else False
    admin_proxy_path = hconfig(ConfigEnum.proxy_path_admin)
    client_proxy_path = hconfig(ConfigEnum.proxy_path_client)
    deprecated_path = hconfig(ConfigEnum.proxy_path)
    if proxy_path == deprecated_path:
        return

    if proxy_path not in [admin_proxy_path, deprecated_path, client_proxy_path]:
        apiflask_abort(400, 'invalid request')

    if is_admin_panel_call() and proxy_path != admin_proxy_path:
        apiflask_abort(400, 'invalid request')
    if is_user_panel_call() and proxy_path != client_proxy_path:
        apiflask_abort(400, 'invalid request')

    if is_api_call(request.path):
        if __is_admin_api_call() and proxy_path != admin_proxy_path:
            return apiflask_abort(400, Markup(f"Invalid Proxy Path <a href=/{admin_proxy_path}/admin>Admin Panel</a>")) if dbg_mode else apiflask_abort(400, 'invalid request')
        if is_user_api_call() and proxy_path != client_proxy_path:
            return apiflask_abort(400, Markup(f"Invalid Proxy Path <a href=/{client_proxy_path}/admin>User Panel</a>")) if dbg_mode else apiflask_abort(400, 'invalid request')


def list_dir_files(dir_path: str) -> List[str]:
    return sorted([f for f in os.listdir(dir_path) if os.path.isfile(os.path.join(dir_path, f))])


def validate_domain_exist(form, field):
    domain = field.data
    if not domain:
        return
    dip = hutils.network.get_domain_ip(domain)
    if dip is None:
        raise ValidationError(
            _("Domain can not be resolved! there is a problem in your domain"))  # type: ignore


def get_proxy_stats_url():
    proxy_stats_url = f'{request.host_url}{g.proxy_path}/proxy-stats/'
    params = f'hostname={proxy_stats_url}api/&port=443&secret=hiddify'
    return f'{proxy_stats_url}?{params}'


def extract_parent_info_from_url(url) -> Tuple[str | None, str | None, str | None]:
    pattern = r'^https?://([^/]+)/([^/]+)/([^/]+)/.*$'
    match = re.match(pattern, url)

    if match:
        domain = match.group(1)
        proxy_path = match.group(2)
        admin_uuid = match.group(3)
        return domain, proxy_path, admin_uuid
    else:
        return None, None, None
# region not used


# def api_v1_auth(function):
#     def wrapper(*args, **kwargs):
#         a_uuid = kwargs.get('admin_uuid')
#         if not a_uuid or a_uuid != AdminUser.get_super_admin_uuid():
#             apiflask_abort(403, 'invalid request')
#         return function(*args, **kwargs)
#     return wrapper


# def current_account_api_key():
#     return g.account.uuid


# def current_account_user_pass() -> Tuple[str, str]:
#     return g.account.username, g.account.password


# def is_telegram_call() -> bool:
#     if request.endpoint and (request.endpoint == 'tgbot' or request.endpoint == 'send_msg'):
#         return True
#     if request.blueprint and request.blueprint == 'api_v1' and ('tgbot' in request.path or 'send_msg' in request.path):
#         return True
#     if '/tgbot/' in request.path or 'send_msg/' in request.path:
#         return True
#     return False


# def is_admin_home_call() -> bool:
#     admin_home = f'{request.host}/{hconfig(ConfigEnum.proxy_path_admin)}/admin/'
#     if f'{request.host}{request.path}' == admin_home:
#         return True
#     return False


# def asset_url(path) -> str:
#     return f"/{g.proxy_path}/{path}"
# endregion



from typing import List
from strenum import StrEnum
from flask import current_app
import subprocess
import os


class Command(StrEnum):
    apply = 'apply'
    install = 'install'
    # reinstall = 'reinstall'
    update = 'update'
    status = 'status'
    restart_services = 'restart-services'
    temporary_short_link = 'temporary-short-link'
    temporary_access = 'temporary-access'
    update_usage = 'update-usage'
    get_cert = 'get-cert'
    apply_users = 'apply-users'
    update_wg_usage = 'update-wg-usage'


def commander(command: Command, run_in_background=True, **kwargs: str | int) -> str | None:
    """
    Run the commander based on the given command type.
    Args:
        command: The type of command to run.
        run_in_background: Whether to run the command in the background.
        **kwargs: Additional arguments to pass to the commander. Accepts the following:
                  url, slug, period for the temporary-short-link command.
                  port for the temporary-access command.
                  domain for the get-cert command
    """
    base_cmd: List[str] = [
        'sudo',
        os.path.join(
            current_app.config['HIDDIFY_CONFIG_PATH'], 'common/commander.py')
    ]

    if command == Command.apply:
        base_cmd.append('apply')
    elif command == Command.install:
        base_cmd.append('install')
    elif command == Command.update:
        base_cmd.append('update')
    elif command == Command.status:
        base_cmd.append('status')
    elif command == Command.restart_services:
        base_cmd.append('restart-services')
    elif command == Command.apply_users:
        base_cmd.append('apply-users')
    elif command == Command.temporary_short_link:
        url = str(kwargs.get('url', ''))
        slug = str(kwargs.get('slug', ''))
        period = kwargs.get('period', '')

        if not url or not slug:
            raise Exception("Invalid input passed to the run_commander function for temporary-short-link command")

        base_cmd.append('temporary-short-link')
        base_cmd.extend(['--url', url, '--slug', slug])
        if period:
            base_cmd.extend(['--period', str(period)])
    elif command == Command.temporary_access:
        port = str(kwargs.get('port'))
        if not port or not port.isnumeric():
            raise Exception("Invalid input passed to the run_commander function for temporary-access command")

        base_cmd.append('temporary-access')
        base_cmd.extend(['--port', port])
    elif command == Command.update_usage:
        base_cmd.append('update-usage')
    elif command == Command.get_cert:
        domain = str(kwargs.get('domain'))
        if not domain:
            raise Exception("Invalid input passed to the run_commander function for get-cert command")
        base_cmd.extend(['get-cert', '--domain', domain])
    elif command == Command.update_wg_usage:
        base_cmd.append('update-wg-usage')
    else:
        raise Exception('WTF is happening!')
    if run_in_background:
        subprocess.Popen(base_cmd, cwd=str(current_app.config['HIDDIFY_CONFIG_PATH']), start_new_session=True)
    else:
        return subprocess.check_output(base_cmd, cwd=str(current_app.config['HIDDIFY_CONFIG_PATH'])).decode()



from redis_cache import RedisCache, CacheDecorator, compact_dump, chunks
from redis_cache import loads as redis_cache_loads
import redis
from pickle import dumps, loads
from loguru import logger

redis_client = redis.from_url('unix:///opt/hiddify-manager/other/redis/run.sock?db=0')


class CustomRedisCache(RedisCache):

    def invalidate_all_cached_functions(self):
        try:
            logger.trace("Invalidating all cached functions")
            chunks_gen = chunks(f'{self.prefix}*', 500)
            for keys in chunks_gen:
                self.client.delete(*keys)
            logger.trace("Successfully invalidated all cached functions")
            return True
        except Exception as err:
            with logger.contextualize(error=err):
                logger.error("Failed to invalidate all cached functions")
            return False


cache = CustomRedisCache(redis_client=redis_client, prefix="h", serializer=dumps, deserializer=loads)


# cache = RedisCache(redis_client=redis_client, exception_handler=exception_handler)
# cache = RedisCache(redis_client=redis_client, prefix="h", serializer=dumps, deserializer=loads, exception_handler=exception_handler)

# def exception_handler(e, original_fn, args, kwargs):
#     print("cache exception occur", e, original_fn, args, kwargs)
#     return original_fn(*args, **kwargs)
#     pass

# class CacheDecorator:
#     def __init__(self, *args, **kwargs):
#         pass

#     def __call__(self, fn):
#         @wraps(fn)
#         def inner(*args, **kwargs):

#             parsed_result = fn(*args, **kwargs)

#             return parsed_result

#         inner.invalidate = self.invalidate
#         inner.invalidate_all = self.invalidate_all
#         inner.instance = self
#         return inner

#     def invalidate(self, *args, **kwargs):
#         pass

#     def invalidate_all(self, *args, **kwargs):
#         pass


# class DisableCache:
#     cache = CacheDecorator


# cache = DisableCache()
# try:
#     @cache.cache()
#     def test():
#         return 1
#     test()
# except Exception as e:
#     import sys
#     print('Caching Error! Disabling cache', e, file=sys.stderr)
#     # cache = DisableCache()



import yaml
from hiddifypanel.models import Proxy, ProxyCDN, ProxyL3, ProxyProto, ProxyTransport, Domain
from hiddifypanel import hutils


def get_clash_config_names(meta_or_normal, domains: list[Domain]):
    allp = []
    for pinfo in hutils.proxy.get_valid_proxies(domains):
        clash = to_clash(pinfo, meta_or_normal)
        if 'msg' not in clash:
            allp.append(clash['name'])

    return yaml.dump(allp, sort_keys=False)


def get_all_clash_configs(meta_or_normal, domains: list[Domain]):
    allp = []
    for pinfo in hutils.proxy.get_valid_proxies(domains):
        clash = to_clash(pinfo, meta_or_normal)
        if 'msg' not in clash:
            allp.append(clash)

    return yaml.dump({"proxies": allp}, sort_keys=False)

# def to_clash_yml(proxy):
#     return yaml.dump(to_clash(proxy,'normal'))


def to_clash(proxy, meta_or_normal):

    name = proxy['name']
    if proxy['l3'] == "kcp":
        return {'name': name, 'msg': "clash does not support kcp", 'type': 'debug'}
    if proxy['proto'] == "ssh":
        return {'name': name, 'msg': "clash does not support ssh", 'type': 'debug'}
    if meta_or_normal == "normal":
        if proxy['proto'] in ["vless", 'tuic', 'hysteria2']:
            return {'name': name, 'msg': f"{proxy['proto']} not supported in clash", 'type': 'debug'}
        if proxy.get('flow'):
            return {'name': name, 'msg': "xtls not supported in clash", 'type': 'debug'}
        if proxy['transport'] == "shadowtls":
            return {'name': name, 'msg': "shadowtls not supported in clash", 'type': 'debug'}
    if proxy['l3'] == ProxyL3.tls_h2 and proxy['proto'] in [ProxyProto.vmess, ProxyProto.vless] and proxy['dbe'].cdn == ProxyCDN.direct:
        return {'name': name, 'msg': "bug tls_h2 vmess and vless in clash meta", 'type': 'warning'}
    base = {}
    # vmess ws
    base["name"] = f"""{proxy['extra_info']} {proxy["name"]} § {proxy['port']} {proxy["dbdomain"].id}"""
    base["type"] = str(proxy["proto"])
    base["server"] = proxy["server"]
    base["port"] = proxy["port"]
    base['alpn'] = proxy['alpn'].split(',')
    if proxy["proto"] == "ssr":
        base["cipher"] = proxy["cipher"]
        base["password"] = proxy["uuid"]
        base["udp"] = True
        base["obfs"] = proxy["ssr-obfs"]
        base["protocol"] = proxy["ssr-protocol"]
        base["obfs-param"] = proxy["fakedomain"]
        return base
    elif proxy["proto"] == "tuic":
        base["uuid"] = proxy["uuid"]
        base["password"] = proxy["uuid"]
        base["disable-sni"] = proxy['allow_insecure']
        base["reduce-rtt"] = True
        base["request-timeout"] = 8000
        base["udp-relay-mode"] = 'native'
        base["congestion-controller"] = 'cubic'
        base['sni'] = proxy['sni']
        return base
    elif proxy["proto"] in ["ss", "v2ray"]:
        base["cipher"] = proxy["cipher"]
        base["password"] = proxy["password"]
        base["udp_over_tcp"] = True
        if proxy["transport"] == "faketls":
            base["plugin"] = "obfs"
            base["plugin-opts"] = {
                "mode": 'tls',
                "host": proxy["fakedomain"]
            }
        elif proxy["transport"] == "shadowtls":
            base["plugin"] = "shadow-tls"
            base["plugin-opts"] = {
                "host": proxy["fakedomain"],
                "password": proxy["proxy_path"],
                "version": 3  # support 1/2/3

            }

        elif proxy["proto"] == "v2ray":
            base["plugin"] = "v2ray-plugin"
            base["type"] = "ss"
            base["plugin-opts"] = {
                "mode": "websocket",
                "tls": "tls" in proxy["l3"],
                "skip-cert-verify": proxy["mode"] == "Fake" or proxy['allow_insecure'],
                "host": proxy['sni'],
                "path": proxy["path"]
            }
        return base
    elif proxy["proto"] == "trojan":
        base["password"] = proxy["uuid"]
        base["sni"] = proxy["sni"]

    else:
        base["uuid"] = proxy["uuid"]
        base["servername"] = proxy["sni"]
        base["tls"] = "tls" in proxy["l3"] or "reality" in proxy["l3"]
    if meta_or_normal == "meta":
        base['client-fingerprint'] = proxy['fingerprint']
    if proxy.get('flow'):
        base["flow"] = proxy['flow']
        # base["flow-show"] = True

    if proxy["proto"] == "vmess":
        base["alterId"] = 0
        base["cipher"] = proxy["cipher"]
    base["udp"] = True

    base["skip-cert-verify"] = proxy["mode"] == "Fake"

    base["network"] = proxy["transport"]

    if base["network"] == "ws":
        base["ws-opts"] = {
            "path": proxy["path"]
        }
        if "host" in proxy:
            base["ws-opts"]["headers"] = {"Host": proxy["host"]}

    if base["network"] == "tcp" and proxy['alpn'] != 'h2':
        if proxy['transport'] != ProxyTransport.XTLS:
            base["network"] = "http"

        if "path" in proxy:
            base["http-opts"] = {
                "path": [proxy["path"]]
            }
            if 'host' in proxy:
                base["http-opts"]["host"] = [proxy["host"]]
    if base["network"] == "tcp" and proxy['alpn'] == 'h2':
        base["network"] = "h2"

        if "path" in proxy:
            base["h2-opts"] = {
                "path": proxy["path"]
            }
            if 'host' in proxy:
                base["h2-opts"]["host"] = [proxy["host"]]
    if base["network"] == "grpc":
        base["grpc-opts"] = {
            "grpc-service-name": proxy["grpc_service_name"]
        }
    if proxy['l3'] == ProxyL3.reality:
        base["reality-opts"] = {
            "public-key": proxy['reality_pbk'],
            "short-id": proxy['reality_short_id'],
        }
        if proxy["transport"] != 'grpc':
            base["network"] = 'tcp'

    return base



import subprocess
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import x25519, ed25519


def get_ed25519_private_public_pair():
    privkey = ed25519.Ed25519PrivateKey.generate()
    pubkey = privkey.public_key()
    priv_bytes = privkey.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.OpenSSH,
        encryption_algorithm=serialization.NoEncryption(),
    )
    pub_bytes = pubkey.public_bytes(
        encoding=serialization.Encoding.OpenSSH,
        format=serialization.PublicFormat.OpenSSH,
    )
    return priv_bytes.decode(), pub_bytes.decode()


def get_wg_private_public_psk_pair():
    try:
        private_key = subprocess.run(["wg", "genkey"], capture_output=True, text=True, check=True).stdout.strip()
        public_key = subprocess.run(["wg", "pubkey"], input=private_key, capture_output=True, text=True, check=True).stdout.strip()
        psk = subprocess.run(["wg", "genpsk"], capture_output=True, text=True, check=True).stdout.strip()
        return private_key, public_key, psk
    except subprocess.CalledProcessError as e:
        print(f"Error: {e}")
        return None, None, None


def generate_x25519_keys():
    priv = x25519.X25519PrivateKey.generate()
    pub = priv.public_key()
    priv_bytes = priv.private_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PrivateFormat.Raw,
        encryption_algorithm=serialization.NoEncryption()
    )
    pub_bytes = pub.public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw
    )
    import base64
    pub_str = base64.urlsafe_b64encode(pub_bytes).decode()[:-1]
    priv_str = base64.urlsafe_b64encode(priv_bytes).decode()[:-1]

    return {'private_key': priv_str, 'public_key': pub_str}



from flask_babel import gettext as _
from typing import List, Union
from flask import request
import maxminddb
import random
import os
import re
import sys
from hiddifypanel.cache import cache
from hiddifypanel.models.config import hconfig
from hiddifypanel.models.config_enum import ConfigEnum
from hiddifypanel import hutils

DEFAULT_IPs = """
mci.ircf.space		MCI
mcix.ircf.space		MCI
mtn.ircf.space		MTN
mtnx.ircf.space		MTN
mkh.ircf.space		MKH
mkhx.ircf.space		MKH
rtl.ircf.space		RTL
hwb.ircf.space		HWB
ast.ircf.space		AST
sht.ircf.space		SHT
prs.ircf.space		PRS
mbt.ircf.space		MBT
ask.ircf.space		ASK
rsp.ircf.space		RSP
afn.ircf.space		AFN
ztl.ircf.space		ZTL
psm.ircf.space		PSM
arx.ircf.space		ARX
smt.ircf.space		SMT
fnv.ircf.space		FNV
dbn.ircf.space		DBN
apt.ircf.space		APT
"""

try:
    IPASN = maxminddb.open_database('GeoLite2-ASN.mmdb') if os.path.exists('GeoLite2-ASN.mmdb') else {}
    IPCOUNTRY = maxminddb.open_database('GeoLite2-Country.mmdb') if os.path.exists('GeoLite2-Country.mmdb') else {}
    __ipcity = maxminddb.open_database('GeoLite2-City.mmdb') if os.path.exists('GeoLite2-City.mmdb') else {}
except Exception as e:
    print("Error can not load maxminddb", file=sys.stderr)
    IPASN = {}
    IPCOUNTRY = {}

__asn_map = {
    '58224': 'MKH',
    '197207': 'MCI',
    '12880': 'ITC',
    '44244': 'MTN',
    '57218': 'RTL',
    '16322': 'PRS',
    '56402': 'HWB',
    '41689': 'AST',
    '43754': 'AST',
    '31549': 'SHT',
    '205647': 'SHT',
    '50810': 'MBT',
    '39308': 'ASK',
    '205207': 'RSP',
    '25184': 'AFR',
    '394510': 'ZTL',
    '206065': 'ZTL',
    '49100': 'PSM'
}


def get_asn_short_name(user_ip: str = '') -> str:
    return __get_asn_short_name_imp(user_ip or get_real_user_ip())


@cache.cache()
def __get_asn_short_name_imp(user_ip: str) -> str:
    try:
        asn_id = get_asn_id(user_ip)
        return __asn_map.get(str(asn_id), "unknown")
    except BaseException:
        return "unknown"


def get_asn_id(user_ip: str = '') -> str:
    return __get_asn_id_imp(user_ip or get_real_user_ip())


@cache.cache()
def __get_asn_id_imp(user_ip: str) -> str:
    try:
        asnres = IPASN.get(user_ip)
        return asnres['autonomous_system_number']
    except BaseException:
        return "unknown"


def get_country(user_ip: str = '') -> Union[dict, str]:
    try:
        user_ip = user_ip or get_real_user_ip()
        return (IPCOUNTRY.get(user_ip) or {}).get('country', {}).get('iso_code', 'unknown')
    except BaseException:
        return 'unknown'


def get_city(user_ip: str = '') -> Union[dict, str]:
    try:
        user_ip = user_ip or get_real_user_ip()
        res = __ipcity.get(user_ip)
        return {'city': res.get('city').get('name'), 'latitude': res.get('latitude'), 'longitude': res.get('longitude'), 'accuracy_radius': res.get('accuracy_radius')}
    except BaseException:
        return 'unknown'


def get_real_user_ip_debug(user_ip: str = '') -> str:
    return __get_real_user_ip_debug_imp(user_ip or get_real_user_ip())


@cache.cache()
def __get_real_user_ip_debug_imp(user_ip) -> str:
    if type(user_ip) is str and ',' in user_ip:
        user_ip = user_ip.split(',')[0]
    asnres = IPASN.get(user_ip) or {}
    asn = f"{asnres.get('autonomous_system_number','unknown')}" if asnres else "unknown"
    asn_dscr = f"{asnres.get('autonomous_system_organization','unknown')}" if asnres else "unknown"
    asn_short = get_asn_short_name(user_ip)
    country = get_country(user_ip)
    default = __get_host_base_on_asn(DEFAULT_IPs, asn_short).replace(".ircf.space", "")
    return f'{user_ip} {country} {asn} {asn_short} {"ERROR" if asn_short=="unknown" else ""} fullname={asn_dscr} default:{default}'


def get_real_user_ip() -> str:
    user_ip = request.remote_addr
    for header in ['CF-Connecting-IP', 'ar-real-ip', 'X-Forwarded-For', "X-Real-IP"]:
        if header in request.headers:
            user_ip = request.headers.get(header)
            break

    return str(user_ip)


def __get_host_base_on_asn(ips: Union[str, List[str]], asn_short: str) -> str:
    if type(ips) == str:
        ips = re.split('[ \t\r\n;,]+', ips.strip())
    valid_hosts = [ip for ip in ips if len(ip) > 5]

    if len(ips) % 2 != 0 or len(valid_hosts) == 0:
        hutils.flask.flash(_("Error! auto cdn ip can not be find, please contact admin."))
        if len(valid_hosts) == 0:
            return ''

    all_hosts = []
    for i in range(0, len(ips), 2):
        if asn_short == ips[i + 1]:
            # print("selected ",ips[i],ips[i+1])
            all_hosts.append(ips[i])

    selected = random.sample(valid_hosts, 1)[0]
    if len(all_hosts):
        selected = random.sample(all_hosts, 1)[0]

    return selected


def get_clean_ip(ips: Union[str, List[str]], resolve: bool = False, default_asn: str = '') -> str:
    if not ips:
        ips = DEFAULT_IPs

    ips = re.split('[ \t\r\n;,]+', ips.strip())
    user_ip = get_real_user_ip()
    asn_short = get_asn_short_name(user_ip)
    country = get_country(user_ip)
    # print("Real user ip",get_real_user_ip_debug(), user_ip,asn_short,country)
    is_morteza_format = any([format for format in __asn_map.values() if format in ips])
    # print("IPs",ips)
    if is_morteza_format:
        if str(country).lower() != hconfig(ConfigEnum.country) and default_asn:
            asn_short = default_asn
        selected_server = __get_host_base_on_asn(ips, asn_short)
    else:
        selected_server = random.sample(ips, 1)[0]
    # print("selected_server",selected_server)
    if resolve:
        selected_server = hutils.network.get_domain_ip(selected_server) or selected_server
    return str(selected_server)



import sqlite3
import json
from typing import Any, Dict, List, Tuple
import uuid as uuid_mod
from datetime import datetime
from dateutil.relativedelta import relativedelta
from hiddifypanel import hutils
from hiddifypanel.models import *
from hiddifypanel.database import db, db_execute
import os
from sqlalchemy import text


def __query_fetch_json(db, query: str, **kwargs) -> List[Dict[str, Any]]:
    try:

        db_execute(query, commit=True, return_val=False, **kwargs)
        r = [dict((db.description[i][0], value)
                  for i, value in enumerate(row)) for row in db.fetchall()]
        return r
    except Exception as err:
        raise err


def __is_reality_domain_exists(reality_domains: Dict[str, Dict[str, Any]], domain: str, network: str) -> bool:
    return domain in reality_domains and network in reality_domains[domain]


def __get_client_current_usage(db, client_email: str) -> Tuple[int, int]:
    r = __query_fetch_json(db, 'select up,down from client_traffics where email = (?)', (client_email,))
    if not r or not r[0] or not r[0]['up'] or not r[0]['down']:
        return 0, 0
    return int(r[0]['up']), int(r[0]['down'])


def __get_users(db, x_ui_inbounds):
    users = {}
    user_count = 1

    for inbound in x_ui_inbounds:
        inbound['settings'] = json.loads(inbound['settings'])
        for c in inbound['settings']['clients']:
            id = c.get('id', c.get('subId')) or 0

            if c.get('enable', False) and id not in users:
                up_current_usage, down_current_usage = __get_client_current_usage(db, c.get('email'))
                current_usage = up_current_usage + down_current_usage

                users[id] = {
                    'name': c.get('email') or f'Imported ({user_count})',
                    'expiry_time': c.get('expiryTime') or 0,
                    'max_usage_bytes': c.get('totalGB') or 0,
                    'telegram_id': c.get('tgId') or 0,
                    'current_usage_bytes': current_usage,
                    'enable': c.get('enable') or None,
                }
            user_count += 1

    return users


def __create_hiddify_user_from_xui_values(id: str, values: Dict[str, Any]) -> "User":
    user = User()
    user.name = values['name']
    user.uuid = id if hutils.auth.is_uuid_valid(id, 4) else uuid_mod.uuid4()

    if str(values['expiry_time']) == '0':
        user.expiry_time = datetime.now() + relativedelta(years=10)
    else:
        user.expiry_time = datetime.fromtimestamp(values['expiry_time'] / 1000)

    user.usage_limit = values['max_usage_bytes']
    user.current_usage = values['current_usage_bytes']
    user.telegram_id = int(values['telegram_id'])
    user.enable = values['enable']
    user.comment = "Imported from X-UI"
    user.admin = AdminUser.current_admin_or_owner()

    return user


def __get_reality_domains(x_ui_inbounds):
    reality_domains = {}

    for inbound in x_ui_inbounds:
        inbound['stream_settings'] = json.loads(inbound['stream_settings'])
        if inbound['stream_settings']['security'] == 'reality':
            d = inbound['stream_settings']['realitySettings']['dest']
            network = inbound['stream_settings']['network']

            if not __is_reality_domain_exists(reality_domains, d, network):
                reality_domains[d] = {'network': network}

    return reality_domains


def __create_hiddify_domain_from_xui_values(domain: str, values: Dict[str, Any]) -> Domain:
    d = Domain()
    d.domain = domain
    d.grpc = values['network'] == 'grpc'
    d.mode = DomainType.reality
    return d


def __extract_users_and_domains_from_xui_db(db_path: str) -> Tuple[List["User"], List["Domain"]]:
    users = {}
    reality_domains = {}

    try:
        with sqlite3.connect(db_path) as conn:
            db = conn.cursor()

            x_ui_inbounds = __query_fetch_json(db, 'select * from inbounds')

            # get xui users and reality domains as a dict
            users = __get_users(db, x_ui_inbounds)
            reality_domains = __get_reality_domains(x_ui_inbounds)

        # convert xui users and reality domains to hiddify users and domains (to hiddify models)
        hiddify_users = [__create_hiddify_user_from_xui_values(id, values) for id, values in users.items()]
        # set default package days
        for u in hiddify_users:
            if not u.package_days:
                u.package_days = 90
            if not u.last_reset_time:
                u.last_reset_time = datetime.now()
            if not u.mode:
                u.mode = UserMode.monthly
        hiddify_reality_domains = [__create_hiddify_domain_from_xui_values(d, values) for d, values in reality_domains.items()]
        return hiddify_users, hiddify_reality_domains

    except Exception as err:
        raise err


def import_data(db_path: str):
    assert os.path.exists(db_path), "The db file doesn't exist"
    try:
        hiddify_users, hiddify_reality_domains = __extract_users_and_domains_from_xui_db(db_path)

        hiddify_users_dict = [u.to_dict() for u in hiddify_users]
        hiddify_domains_dict = [d.to_dict() for d in hiddify_reality_domains]

        for u in hiddify_users_dict:
            User.add_or_update(commit=False, **u)

        for d in hiddify_domains_dict:
            from hiddifypanel.panel import hiddify
            hiddify.add_or_update_domain(commit=False, **d)

        db.session.commit()  # type: ignore
    except Exception as err:
        raise err



import datetime

from flask_admin.contrib.sqla import ModelView
from flask_babel import lazy_gettext as _
from wtforms import TextAreaField
from wtforms.fields import IntegerField, SelectField, DecimalField
from wtforms.widgets import TextArea

from hiddifypanel.models import *


# from gettext import gettext as _


class DaysLeftField(IntegerField):
    def process_data(self, value):
        if value is not None:
            days_left = (value - datetime.date.today()).days
            self.data = days_left
        else:
            self.data = None

    def process_formdata(self, valuelist):
        if valuelist and valuelist[0]:
            days_left = valuelist[0]
            new_date_value = datetime.date.today() + datetime.timedelta(days=int(days_left))
            self.data = new_date_value
        else:
            self.data = None


class LastResetField(IntegerField):
    def process_data(self, value):
        if value is not None:
            days_left = (datetime.date.today() - value).days
            self.data = days_left
        else:
            self.data = None

    def process_formdata(self, valuelist):
        if valuelist and valuelist[0]:
            days_left = valuelist[0]
            new_date_value = datetime.date.today() - datetime.timedelta(days=int(days_left))
            self.data = new_date_value
        else:
            self.data = None


class CKTextAreaWidget(TextArea):
    def __call__(self, field, **kwargs):
        if kwargs.get('class'):
            kwargs['class'] += ' ckeditor'
        else:
            kwargs.setdefault('class', 'ckeditor')
        return super(CKTextAreaWidget, self).__call__(field, **kwargs)


class CKTextAreaField(TextAreaField):
    widget = CKTextAreaWidget()


class MessageAdmin(ModelView):
    extra_js = ['//cdn.ckeditor.com/4.6.0/standard/ckeditor.js']

    form_overrides = {
        'body': CKTextAreaField
    }


class EnumSelectField(SelectField):
    def __init__(self, enum, *args, **kwargs):
        choices = [(str(enum_value.value), _(enum_value.name)) for enum_value in enum]
        super().__init__(*args, choices=choices, **kwargs)


class UsageField(DecimalField):
    def process_data(self, value):
        if value is not None:
            self.data = value / ONE_GIG
        else:
            self.data = None

    def process_formdata(self, valuelist):

        if valuelist and valuelist[0]:
            self.data = int(float(valuelist[0]) * ONE_GIG)
        else:
            self.data = None
